# coding=utf-8
from __future__ import unicode_literals
import os
import sys
import platform

try:
    import ConfigParser as configparser
except ImportError:
    import configparser


# download delegate
__wget__ = ('WGET_PATH',)
__thunder__ = ('XUNLEI_LX_PATH',)
__transmission__ = ('TRANSMISSION_RPC_URL', 'TRANSMISSION_RPC_PORT',)
__aria2__ = ('ARIA2_RPC_URL', 'ARIA2_RPC_TOKEN',)

__download_delegate__ = __wget__ + __thunder__ + __aria2__ + __transmission__

# fake __all__
__all__ = ('BANGUMI_MOE_URL', 'BGMI_SAVE_PATH', 'DOWNLOAD_DELEGATE', 'MAX_PAGE',
           'DATA_SOURCE', 'SUPPORT_WEBSITE', 'BGMI_TMP_PATH', 'DANMAKU_API_URL', 'LANG',)

# cannot be rewrite
__readonly__ = ('BGMI_PATH', 'DB_PATH', 'CONFIG_FILE_PATH',
                'SUPPORT_WEBSITE', 'IS_PYTHON3', 'SCRIPT_PATH',
                'SCRIPT_DB_PATH',)

# writeable
__writeable__ = tuple([i for i in __all__ if i not in __readonly__])

# the real __all__
__all__ = __all__ + __download_delegate__ + __readonly__


download_delegate_map = {
    'rr!': __wget__,
    'aria2-rpc': __aria2__,
    'xunlei': __thunder__,
    'transmission-rpc': __transmission__,
}


# --------- Immutable ---------- #
if platform.system() == 'Windows':
    BGMI_PATH = os.path.join(os.environ.get('USERPROFILE', None), '.bgmi')
    if not BGMI_PATH:
        raise SystemExit
else:
    BGMI_PATH = os.path.join(os.environ.get('HOME', '/tmp'), '.bgmi')

DB_PATH = os.path.join(BGMI_PATH, 'bangumi.db')
CONFIG_FILE_PATH = os.path.join(BGMI_PATH, 'bgmi.cfg')

SCRIPT_DB_PATH = os.path.join(BGMI_PATH, 'script.db')
SCRIPT_PATH = os.path.join(BGMI_PATH, 'scripts')


def read_config():
    c = configparser.ConfigParser()
    if not os.path.exists(CONFIG_FILE_PATH):
        write_default_config()
        return
    c.read(CONFIG_FILE_PATH)

    for i in __writeable__:
        if c.has_option('bgmi', i):
            globals().update({i: c.get('bgmi', i)})

    for i in download_delegate_map.get(DOWNLOAD_DELEGATE):
        if c.has_option(DOWNLOAD_DELEGATE, i):
            globals().update({i: c.get(DOWNLOAD_DELEGATE, i)})


def print_config():
    c = configparser.ConfigParser()
    if not os.path.exists(CONFIG_FILE_PATH):
        return

    c.read(CONFIG_FILE_PATH)
    print('[bgmi]')
    for i in __writeable__:
        print('{0}={1}'.format(i, c.get('bgmi', i)))

    print('\n[{0}]'.format(DOWNLOAD_DELEGATE))
    for i in download_delegate_map.get(DOWNLOAD_DELEGATE):
        print('{0}={1}'.format(i, c.get(DOWNLOAD_DELEGATE, i)))


def write_default_config():
    c = configparser.ConfigParser()
    if not c.has_section('bgmi'):
        c.add_section('bgmi')

    for i in __writeable__:
        v = globals().get(i, None)
        c.set('bgmi', i, v)

    if DOWNLOAD_DELEGATE not in download_delegate_map.keys():
        raise Exception(DOWNLOAD_DELEGATE)

    if not c.has_section(DOWNLOAD_DELEGATE):
        c.add_section(DOWNLOAD_DELEGATE)

    for i in download_delegate_map.get(DOWNLOAD_DELEGATE):
        v = globals().get(i, None)
        c.set(DOWNLOAD_DELEGATE, i, v)

    try:
        c.write(open(CONFIG_FILE_PATH, 'w'))
    except IOError:
        print('[-] Error writing to config file and ignored')


def write_config(config=None, value=None):
    if not os.path.exists(CONFIG_FILE_PATH):
        write_default_config()

    c = configparser.ConfigParser()
    c.read(CONFIG_FILE_PATH)

    if config is not None and config not in __writeable__ and config not in __download_delegate__:
        print('{0} does not exist or not writeable'.format(config))
        exit(1)

    try:
        if config is None and value is None:
            print_config()
        elif value is None:
            if config in __download_delegate__:
                print('{0}={1}'.format(config, c.get(DOWNLOAD_DELEGATE, config)))
            else:
                print('{0}={1}'.format(config, c.get('bgmi', config)))
        else:
            if config in __writeable__:
                c.set('bgmi', config, value)
                with open(CONFIG_FILE_PATH, 'w') as f:
                    c.write(f)
                read_config()

            if config == 'DOWNLOAD_DELEGATE':
                if not c.has_section(DOWNLOAD_DELEGATE):
                    c.add_section(DOWNLOAD_DELEGATE)
                    for i in download_delegate_map.get(DOWNLOAD_DELEGATE):
                        v = globals().get(i, '')
                        c.set(DOWNLOAD_DELEGATE, i, v)

                    with open(CONFIG_FILE_PATH, 'w') as f:
                        c.write(f)

            if config in download_delegate_map.get(DOWNLOAD_DELEGATE):
                c.set(DOWNLOAD_DELEGATE, config, value)
                with open(CONFIG_FILE_PATH, 'w') as f:
                    c.write(f)

            print_config()

    except configparser.NoOptionError:
        write_default_config()


# --------- Writeable ---------- #
# Setting bangumi.moe url
BANGUMI_MOE_URL = 'https://bangumi_moe.ricterz.me'

# Setting bangumi.moe url
DATA_SOURCE = 'bangumi_moe'
SUPPORT_WEBSITE = [
    {
        'view': '萌番组 https://bangumi.moe/',
        'id': "bangumi_moe",
        'url': BANGUMI_MOE_URL
    },
    {
        'view': '蜜柑计划 https://mikanani.me/',
        'id': 'mikan_project',
        'url': 'https://mikanani.me/'
    },
]

# BGmi user path
BGMI_SAVE_PATH = os.path.join(BGMI_PATH, 'bangumi')

# Xunlei offline download
XUNLEI_LX_PATH = os.path.join(BGMI_PATH, 'bgmi-lx')

# temp path
BGMI_TMP_PATH = os.path.join(BGMI_PATH, 'tmp')

# Download delegate
DOWNLOAD_DELEGATE = 'aria2-rpc'

# danmaku api url, https://github.com/DIYgod/DPlayer#related-projects
DANMAKU_API_URL = ''

# bangumi cover url
COVER_URL = 'https://bangumi_moe.ricterz.me'

# language
LANG = 'zh_cn'

# max page
MAX_PAGE = '3'

# aria2
ARIA2_RPC_URL = 'http://localhost:6800/rpc'
ARIA2_RPC_TOKEN = 'token:'

# path of wget
WGET_PATH = '/usr/bin/wget'

# transmission-rpc
TRANSMISSION_RPC_URL = '127.0.0.1'
TRANSMISSION_RPC_PORT = '9091'

# tag of bangumi on bangumi.moe
BANGUMI_TAG = '549ef207fe682f7549f1ea90'


# ------------------------------ #
# !!! Read config from file and write to globals() !!!
read_config()
# ------------------------------ #


# --------- Read-Only ---------- #
# Python version
IS_PYTHON3 = sys.version_info > (3, 0)

# Detail URL



# coding=utf-8
from __future__ import print_function, unicode_literals

import datetime
import os
import re
import string
import time
from collections import defaultdict
from itertools import chain
import bgmi.config
from bgmi.config import MAX_PAGE
from bgmi.models import Bangumi, Filter, Subtitle, STATUS_FOLLOWED, STATUS_UPDATED
from bgmi.script import ScriptRunner
from bgmi.utils import (print_warning, print_info,
                        test_connection, get_terminal_col, GREEN, YELLOW, COLOR_END)

if bgmi.config.IS_PYTHON3:
    _unicode = str
else:
    _unicode = unicode


FETCH_EPISODE_ZH = re.compile("第?\s?(\d{1,3})\s?[話话集]")
FETCH_EPISODE_WITH_BRACKETS = re.compile('[【\[](\d+)\s?(?:END)?[】\]]')
FETCH_EPISODE_ONLY_NUM = re.compile('^([\d]{2,})$')
FETCH_EPISODE_RANGE = re.compile('[\d]{2,}\s?-\s?([\d]{2,})')
FETCH_EPISODE_OVA_OAD = re.compile('([\d]{2,})\s?\((?:OVA|OAD)\)]')
FETCH_EPISODE_WITH_VERSION = re.compile('[【\[](\d+)\s? *v\d(?:END)?[】\]]')
FETCH_EPISODE = (
    FETCH_EPISODE_ZH, FETCH_EPISODE_WITH_BRACKETS, FETCH_EPISODE_ONLY_NUM,
    FETCH_EPISODE_RANGE,
    FETCH_EPISODE_OVA_OAD, FETCH_EPISODE_WITH_VERSION)


class BaseWebsite(object):
    # todo : 00-11这种识别错误
    @staticmethod
    def parse_episode(episode_title):
        """
        parse episode from title

        :param episode_title: episode title
        :type episode_title: str
        :return: episode of this title
        :rtype: int
        """

        _ = FETCH_EPISODE_ZH.findall(episode_title)
        if _ and _[0].isdigit():
            return int(_[0])

        _ = FETCH_EPISODE_WITH_BRACKETS.findall(episode_title)
        if _ and _[0].isdigit():
            return int(_[0])

        _ = FETCH_EPISODE_WITH_VERSION.findall(episode_title)
        if _ and _[0].isdigit():
            return int(_[0])

        for split_token in ['【', '[', ' ']:
            for i in episode_title.split(split_token):
                for regexp in FETCH_EPISODE:
                    match = regexp.findall(i)
                    if match and match[0].isdigit():
                        return int(match[0])
        return 0

    def search(self, keyword='', count=1, filter_=None):
        if not filter_:
            filter_ = '(.*)'
        match_title = re.compile(filter_)

        result = self.search_by_keyword(keyword, count)
        # filter
        filtered_result = []
        for episode in result:
            if match_title.match(episode['title']):
                filtered_result.append(episode)
        result = filtered_result[:]

        # remove duplicated episode in result
        ret = []
        episodes = list({i['episode'] for i in result})
        for i in result:
            if i['episode'] in episodes:
                ret.append(i)
                del episodes[episodes.index(i['episode'])]

        if os.environ.get('DEBUG', None):
            for i in ret:
                print(i['title'], i['download'])
        return ret

    @staticmethod
    def save_data(data):
        b = Bangumi(**data)
        b.save()

    def fetch(self, save=False, group_by_weekday=True, status=False):
        bangumi_result, subtitle_group_result = self.fetch_bangumi_calendar_and_subtitle_group()
        if subtitle_group_result:
            for subtitle_group in subtitle_group_result:
                s = Subtitle(id=_unicode(subtitle_group['id']), name=_unicode(subtitle_group['name']))
                if not s.select():
                    s.save()
        if not bangumi_result:
            print('no result return None')
            return []

        if save:
            for bangumi in bangumi_result:
                self.save_data(bangumi)
        if group_by_weekday:
            result_group_by_weekday = defaultdict(list)
            for bangumi in bangumi_result:
                result_group_by_weekday[bangumi['update_time'].lower()].append(bangumi)
            bangumi_result = result_group_by_weekday
        return bangumi_result

    def bangumi_calendar(self, force_update=False, today=False, followed=False, save=True):
        env_columns = get_terminal_col()

        col = 42

        if env_columns < col:
            print_warning('terminal window is too small.')
            env_columns = col

        row = int(env_columns / col if env_columns / col <= 3 else 3)

        if force_update and not test_connection():
            force_update = False
            print_warning('network is unreachable')

        if force_update:
            print_info('fetching bangumi info ...')
            Bangumi.delete_all()
            weekly_list = self.fetch(save=save, status=True)
        else:
            if followed:
                weekly_list_followed = Bangumi.get_all_bangumi(status=STATUS_FOLLOWED)
                weekly_list_updated = Bangumi.get_all_bangumi(status=STATUS_UPDATED)
                weekly_list = defaultdict(list)
                for k, v in chain(weekly_list_followed.items(), weekly_list_updated.items()):
                    weekly_list[k].extend(v)
            else:
                weekly_list = Bangumi.get_all_bangumi()

        if not weekly_list:
            if not followed:
                print_warning('warning: no bangumi schedule, fetching ...')
                weekly_list = self.fetch(save=save)
            else:
                print_warning('you have not subscribed any bangumi')

        def shift(seq, n):
            n %= len(seq)
            return seq[n:] + seq[:n]

        def print_line():
            num = col - 3
            split = '-' * num + '   '
            print(split * row)

        if today:
            weekday_order = (Bangumi.week[datetime.datetime.today().weekday()],)
        else:
            weekday_order = shift(Bangumi.week, datetime.datetime.today().weekday())

        runner = ScriptRunner()
        patch_list = runner.get_models_dict()

        for i in patch_list:
            weekly_list[i['update_time'].lower()].append(i)

        spacial_append_chars = ['Ⅱ', 'Ⅲ', '♪', 'Δ', '×', '☆', 'é', '·', '♭']
        spacial_remove_chars = []

        for index, weekday in enumerate(weekday_order):
            if weekly_list[weekday.lower()]:
                print('%s%s. %s' % (GREEN,
                                    weekday if not today else 'Bangumi Schedule for Today (%s)' % weekday, COLOR_END),
                      end='')
                if not followed:
                    print()
                    print_line()

                for i, bangumi in enumerate(weekly_list[weekday.lower()]):
                    if bangumi['status'] in (STATUS_UPDATED, STATUS_FOLLOWED) and 'episode' in bangumi:
                        bangumi['name'] = '%s(%d)' % (bangumi['name'], bangumi['episode'])

                    half = len(re.findall('[%s]' % string.printable, bangumi['name']))
                    full = (len(bangumi['name']) - half)
                    space_count = col - 2 - (full * 2 + half)

                    for s in spacial_append_chars:
                        if s in bangumi['name']:
                            space_count += 1

                    for s in spacial_remove_chars:
                        if s in bangumi['name']:
                            space_count -= 1

                    if bangumi['status'] == STATUS_FOLLOWED:
                        bangumi['name'] = '%s%s%s' % (YELLOW, bangumi['name'], COLOR_END)

                    if bangumi['status'] == STATUS_UPDATED:
                        bangumi['name'] = '%s%s%s' % (GREEN, bangumi['name'], COLOR_END)

                    if followed:
                        if i > 0:
                            print(' ' * 5, end='')
                        f = map(lambda x: x['name'], Subtitle.get_subtitle(bangumi['subtitle_group'].split(', ')))
                        print(bangumi['name'], ', '.join(f))
                    else:
                        print(' ' + bangumi['name'], ' ' * space_count, end='')
                        if (i + 1) % row == 0 or i + 1 == len(weekly_list[weekday.lower()]):
                            print()

                if not followed:
                    print()
                    # print_line()

    def get_maximum_episode(self, bangumi, subtitle=True, ignore_old_row=True, max_page=MAX_PAGE):
        followed_filter_obj = Filter(bangumi_name=bangumi.name)
        followed_filter_obj.select_obj()

        subtitle_group = followed_filter_obj.subtitle if followed_filter_obj and subtitle else None
        include = followed_filter_obj.include if followed_filter_obj and subtitle else None
        exclude = followed_filter_obj.exclude if followed_filter_obj and subtitle else None
        regex = followed_filter_obj.regex if followed_filter_obj and subtitle else None

        data = [i for i in self.fetch_episode(_id=bangumi.keyword, name=bangumi.name,
                                              subtitle_group=subtitle_group,
                                              include=include, exclude=exclude, regex=regex, max=max_page)
                if i['episode'] is not None]

        if ignore_old_row:
            data = [row for row in data if row['time'] > int(time.time()) - 3600 * 24 * 30 * 3]  # three month

        if data:
            bangumi = max(data, key=lambda _i: _i['episode'])
            # pprint(bangumi)
            # pprint(data)
            return bangumi, data
        else:
            return {'episode': 0}, []

    def fetch_episode(self, _id, name='', **kwargs):
        result = []

        subtitle_group = kwargs.get('subtitle_group', None)
        include = kwargs.get('include', None)
        exclude = kwargs.get('exclude', None)
        regex = kwargs.get('regex', None)
        max_page = int(kwargs.get('max', int(MAX_PAGE)))

        if subtitle_group and subtitle_group.split(', '):
            condition = subtitle_group.split(', ')
            response_data = self.fetch_episode_of_bangumi(bangumi_id=_id, subtitle_list=condition)
        else:
            response_data = self.fetch_episode_of_bangumi(bangumi_id=_id, max_page=max_page)

        for info in response_data:
            if '合集' not in info['title']:
                info['name'] = name
                result.append(info)
                # result.append({
                #     'download': info['magnet'],
                #     'name': name,
                #     'subtitle_group': info['team_id'],
                #     'title': info['title'],
                #     'episode': self.parse_episode(info['title']),
                #     'time': int(time.mktime(datetime.datetime.strptime(info['publish_time'].split('.')[0],
                #                                                        "%Y-%m-%dT%H:%M:%S").timetuple()))
                # })

        if include:
            include_list = list(map(lambda s: s.strip(), include.split(',')))
            result = list(filter(lambda s: True if all(map(lambda t: t in s['title'],
                                                           include_list)) else False, result))

        if exclude:
            exclude_list = list(map(lambda s: s.strip(), exclude.split(',')))
            result = list(filter(lambda s: True if all(map(lambda t: t not in s['title'],
                                                           exclude_list)) else False, result))

        if regex:
            try:
                match = re.compile(regex)
                result = list(filter(lambda s: True if match.findall(s['title']) else False, result))
            except re.error:
                pass

        # pprint(result)
        return result

    def search_by_keyword(self, keyword, count):
        return []

    def fetch_bangumi_calendar_and_subtitle_group(self):
        return [], []

    def fetch_episode_of_bangumi(self, bangumi_id, subtitle_list=None, max_page=MAX_PAGE):
        return []



# coding=utf-8

CREATE_TABLE_BANGUMI = '''CREATE TABLE IF NOT EXISTS bangumi (
          id INTEGER PRIMARY KEY AUTOINCREMENT,
          name TEXT NOT NULL UNIQUE,
          subtitle_group TEXT NOT NULL,
          keyword TEXT,
          update_time CHAR(5) NOT NULL,
          cover TEXT,
          status INTEGER DEFAULT 0
        )'''


CREATE_TABLE_FOLLOWED = '''CREATE TABLE IF NOT EXISTS followed (
          id INTEGER PRIMARY KEY AUTOINCREMENT,
          bangumi_name TEXT NOT NULL UNIQUE,
          episode INTEGER DEFAULT 0,
          status INTEGER DEFAULT 1,
          updated_time INTEGER DEFAULT 0
        )'''


CREATE_TABLE_DOWNLOAD = '''CREATE TABLE IF NOT EXISTS download (
          id INTEGER PRIMARY KEY AUTOINCREMENT,
          name TEXT NOT NULL,
          title TEXT NOT NULL,
          episode INTEGER DEFAULT 0,
          download TEXT,
          status INTEGER DEFAULT 0
        )'''


CREATE_TABLE_FOLLOWED_FILTER = '''CREATE TABLE IF NOT EXISTS filter (
          id INTEGER PRIMARY KEY  AUTOINCREMENT,
          bangumi_name TEXT  UNIQUE NOT NULL,
          subtitle TEXT,
          include TEXT,
          exclude TEXT,
          regex TEXT
        )'''


CREATE_TABLE_SUBTITLE = '''CREATE TABLE IF NOT EXISTS subtitle (
          id TEXT PRIMARY KEY UNIQUE NOT NULL,
          name TEXT
        )'''


CREATE_TABLE_SCRIPT = '''CREATE TABLE IF NOT EXISTS scripts (
          id INTEGER PRIMARY KEY  AUTOINCREMENT,
          bangumi_name TEXT UNIQUE NOT NULL,
          episode INTEGER DEFAULT 0,
          status INTEGER DEFAULT 1
        )'''

CLEAR_TABLE_ = 'DELETE  FROM {}'



# coding=utf-8
ACTION_ADD = 'add'
ACTION_FETCH = 'fetch'
ACTION_FILTER = 'filter'
ACTION_DELETE = 'delete'
ACTION_UPDATE = 'update'
ACTION_CAL = 'cal'
ACTION_CONFIG = 'config'
ACTION_DOWNLOAD = 'download'
ACTION_FOLLOWED = 'followed'
ACTION_LIST = 'list'
ACTION_MARK = 'mark'
ACTION_SEARCH = 'search'
ACTION_SOURCE = 'source'
ACTIONS = (ACTION_ADD, ACTION_DELETE, ACTION_UPDATE, ACTION_CAL,
           ACTION_CONFIG, ACTION_FILTER, ACTION_FETCH, ACTION_DOWNLOAD, ACTION_FOLLOWED,
           ACTION_LIST, ACTION_MARK, ACTION_SEARCH, ACTION_SOURCE)

FILTER_CHOICE_TODAY = 'today'
FILTER_CHOICE_ALL = 'all'
FILTER_CHOICE_FOLLOWED = 'followed'
FILTER_CHOICES = (FILTER_CHOICE_ALL, FILTER_CHOICE_FOLLOWED, FILTER_CHOICE_TODAY)


DOWNLOAD_ACTION_LIST = 'list'
DOWNLOAD_ACTION_MARK = 'mark'
DOWNLOAD_ACTION = (DOWNLOAD_ACTION_LIST, DOWNLOAD_ACTION_MARK)


DOWNLOAD_CHOICE_LIST_ALL = 'all'
DOWNLOAD_CHOICE_LIST_NOT_DOWNLOAD = 'not_downloaded'
DOWNLOAD_CHOICE_LIST_DOWNLOADING = 'downloading'
DOWNLOAD_CHOICE_LIST_DOWNLOADED = 'downloaded'
DOWNLOAD_CHOICE_LIST_DICT = {
    None: DOWNLOAD_CHOICE_LIST_ALL,
    0: DOWNLOAD_CHOICE_LIST_NOT_DOWNLOAD,
    1: DOWNLOAD_CHOICE_LIST_DOWNLOADING,
    2: DOWNLOAD_CHOICE_LIST_DOWNLOADED,
}

DOWNLOAD_CHOICE = (DOWNLOAD_CHOICE_LIST_ALL, DOWNLOAD_CHOICE_LIST_DOWNLOADED,
                   DOWNLOAD_CHOICE_LIST_DOWNLOADING, DOWNLOAD_CHOICE_LIST_NOT_DOWNLOAD)

FOLLOWED_ACTION_LIST = 'list'
FOLLOWED_ACTION_MARK = 'mark'
FOLLOWED_CHOICE = (FOLLOWED_ACTION_LIST, FOLLOWED_ACTION_MARK)



# encoding: utf-8
from __future__ import print_function, unicode_literals
import os
import json
import datetime
import hashlib
import tornado.ioloop
import tornado.options
import tornado.httpserver
import tornado.web
import tornado.template
from tornado.options import options, define
from icalendar import Calendar, Event, vText

from collections import OrderedDict, defaultdict
from bgmi import __version__
from bgmi.config import BGMI_SAVE_PATH, DB_PATH, DANMAKU_API_URL
from bgmi.models import Download, Bangumi, Followed, STATUS_NORMAL, STATUS_UPDATING, STATUS_END
from bgmi.fetch import website
from bgmi.script import ScriptRunner

COVER_URL = website.cover_url

WEEK = ('Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat', 'Sun')
define('port', default=8888, help='listen on the port', type=int)
define('address', default='0.0.0.0', help='binding at given address', type=str)


def md5(_, string):
    return hashlib.md5(string.encode('utf-8')).hexdigest()


def make_dicts(cursor, row):
    return dict((cursor.description[idx][0], value)
                for idx, value in enumerate(row))


class BaseHandler(tornado.web.RequestHandler):
    patch_list = None

    def data_received(self, chunk):
        pass

    def __init__(self, *args, **kwargs):

        if self.patch_list is None:
            runner = ScriptRunner()
            self.patch_list = runner.get_models_dict()

        super(BaseHandler, self).__init__(*args, **kwargs)


class BangumiHandler(BaseHandler):
    def get(self, _):
        self.set_header('Content-Type', 'text/html')
        self.write('<h1>BGmi HTTP Service</h1>')
        self.write('<pre>Please modify your web server configure file\n'
                   'to server this path to \'%s\'.\n'
                   'e.g.\n\n'
                   '...\n'
                   'autoindex on;\n'
                   'location /bangumi {\n'
                   '    alias %s;\n'
                   '}\n'
                   '...\n</pre>' % (BGMI_SAVE_PATH, BGMI_SAVE_PATH)
                   )
        self.finish()


class BangumiPlayerHandler(BaseHandler):
    def get(self, bangumi_name):
        data = Followed(bangumi_name=bangumi_name)
        data.select_obj()

        bangumi_obj = Bangumi(name=bangumi_name)
        bangumi_obj.select_obj()

        if not data:
            for i in self.patch_list:
                if bangumi_name == i['bangumi_name']:
                    data = i
                    break

        if not data:
            return self.write_error(404)

        episode_list = {}
        bangumi_path = os.path.join(BGMI_SAVE_PATH, bangumi_name)
        for root, _, files in os.walk(bangumi_path):
            if not _ and files:
                _ = root.replace(bangumi_path, '').split('/')
                base_path = root.replace(BGMI_SAVE_PATH, '')
                if len(_) >= 2:
                    episode_path = root.replace(os.path.join(BGMI_SAVE_PATH, bangumi_name), '')
                    episode = int(episode_path.split('/')[1])
                else:
                    episode = -1

                for bangumi in files:
                    if bangumi.lower().endswith('.mp4'):
                        episode_list[episode] = {'path': os.path.join(base_path, bangumi)}
                        break

        if not episode_list:
            self.write('_(:3 There are nothing to play, please try again later.')
            self.finish()
        else:
            self.render('templates/dplayer.html', bangumi=episode_list, bangumi_name=bangumi_name,
                        bangumi_cover='{}/{}'.format(COVER_URL, bangumi_obj['cover']), DANMAKU_URL=DANMAKU_API_URL)


class ImageCSSHandler(BaseHandler):
    def get(self):
        data = Followed.get_all_followed(status=None, bangumi_status=None)
        for _ in data:
            _['cover'] = '{}/{}'.format(COVER_URL, _['cover'])
        data.extend(self.patch_list)

        self.set_header('Content-Type', 'text/css; charset=utf-8')
        self.render('templates/image.css', data=data)


class RssHandler(BaseHandler):
    def get(self):
        data = Download.get_all_downloads()
        self.set_header('Content-Type', 'text/xml')
        self.render('templates/download.xml', data=data)


class CalendarHandler(BaseHandler):
    def get(self):
        type_ = self.get_argument('type', 0)

        cal = Calendar()
        cal.add('prodid', '-//BGmi Followed Bangumi Calendar//bangumi.ricterz.me//')
        cal.add('version', '2.0')

        data = Followed.get_all_followed(order='followed.updated_time', desc=True)
        data.extend(self.patch_list)

        if type_ == 0:

            bangumi = defaultdict(list)
            [bangumi[Bangumi.week.index(i['update_time']) + 1].append(i['bangumi_name']) for i in data]

            weekday = datetime.datetime.now().weekday()
            for i, k in enumerate(range(weekday, weekday + 7)):
                if k % 7 in bangumi:
                    for v in bangumi[k % 7]:
                        event = Event()
                        event.add('summary', v)
                        event.add('dtstart', datetime.datetime.now().date() + datetime.timedelta(i - 1))
                        event.add('dtend', datetime.datetime.now().date() + datetime.timedelta(i - 1))
                        cal.add_component(event)
        else:
            data = [bangumi for bangumi in data if bangumi['status'] == 2]
            for bangumi in data:
                event = Event()
                event.add('summary', 'Updated: {}'.format(bangumi['bangumi_name']))
                event.add('dtstart', datetime.datetime.now().date())
                event.add('dtend', datetime.datetime.now().date())
                cal.add_component(event)

        cal.add('name', 'Bangumi Calendar')
        cal.add('X-WR-CALNAM', 'Bangumi Calendar')
        cal.add('description', 'Followed Bangumi Calendar')
        cal.add('X-WR-CALDESC', 'Followed Bangumi Calendar')

        self.write(cal.to_ical())
        self.finish()


class MainHandler(BaseHandler):
    def get(self):
        is_json = self.get_argument('json', False)
        is_old = self.get_argument('old', False)

        if not os.path.exists(DB_PATH):
            self.write('BGmi db file not found.')
            self.finish()
            return

        data = Followed.get_all_followed(STATUS_NORMAL, STATUS_UPDATING,
                                         order='followed.updated_time', desc=True)

        followed = list(map(lambda b: b['bangumi_name'], data))
        followed.extend(list(map(lambda b: b['bangumi_name'], self.patch_list)))

        data = Followed.get_all_followed(STATUS_NORMAL, STATUS_UPDATING if not is_old else STATUS_END,
                                         order='followed.updated_time', desc=True)
        data.extend(self.patch_list)
        data.sort(key=lambda _: _['status'])
        data.reverse()

        calendar = Bangumi.get_all_bangumi()

        for i in self.patch_list:
            calendar[i['update_time'].lower()].append(i)

        def shift(seq, n):
            n %= len(seq)
            return seq[n:] + seq[:n]

        weekday_order = shift(WEEK, datetime.datetime.today().weekday())
        cal_ordered = OrderedDict()

        for week in weekday_order:
            cal_ordered[week] = calendar[week.lower()]

        if is_json:
            self.write(json.dumps(cal_ordered))
            self.finish()
        else:
            self.render('templates/bangumi.html', data=data, cal=cal_ordered,
                        followed=list(followed), version=__version__)


def make_app():
    settings = {
        'static_path': os.path.join(os.path.dirname(__file__), 'static'),
        'ui_methods': [{'md5': md5}],
        'debug': True,
    }
    return tornado.web.Application([
        (r'/', MainHandler),
        (r'^/css/image.css$', ImageCSSHandler),
        (r'^/player/(.*)/$', BangumiPlayerHandler),
        (r'^/bangumi/(.*)', BangumiHandler),
        (r'^/rss$', RssHandler),
        (r'^/calendar.ics$', CalendarHandler),
    ], **settings)


def main():
    tornado.options.parse_command_line()
    print('BGmi HTTP Server listening on %s:%d' % (options.address, options.port))
    http_server = tornado.httpserver.HTTPServer(make_app())
    http_server.listen(options.port, address=options.address)
    tornado.ioloop.IOLoop.current().start()


if __name__ == "__main__":
    main()



# coding=utf-8
from __future__ import print_function, unicode_literals

import os
import time

import requests
from bs4 import BeautifulSoup
import bs4

from bgmi.config import LANG, MAX_PAGE
# from bgmi.utils import (print_warning, print_info, print_error)

from bgmi.website.base import BaseWebsite

week = ['Sun', 'Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat']


def process_name(data):
    lang = 'zh_cn' if LANG not in ('zh_cn', 'zh_tw', 'ja', 'en') else LANG
    return {i['_id']: i['locale'][lang] for i in data}


def get_weekly_bangumi():
    r = requests.get('https://mikanani.me/')
    soup = BeautifulSoup(r.text, 'lxml')
    sunday = soup.find('div', attrs={'class': 'sk-bangumi', 'data-dayofweek': "0"})
    monday = soup.find('div', attrs={'class': 'sk-bangumi', 'data-dayofweek': "1"})
    tuesday = soup.find('div', attrs={'class': 'sk-bangumi', 'data-dayofweek': "2"})
    wednesday = soup.find('div', attrs={'class': 'sk-bangumi', 'data-dayofweek': "3"})
    thursday = soup.find('div', attrs={'class': 'sk-bangumi', 'data-dayofweek': "4"})
    friday = soup.find('div', attrs={'class': 'sk-bangumi', 'data-dayofweek': "5"})
    saturday = soup.find('div', attrs={'class': 'sk-bangumi', 'data-dayofweek': "6"})
    return [sunday, monday, tuesday, wednesday, thursday, friday, saturday]


def parser_day_bangumi(soup):
    """

    :param soup:
    :type soup: bs4.Tag
    :return: list
    :rtype: list[dict]
    """
    li = []
    for soup in soup.find_all('li'):

        url = soup.select_one('a')
        span = soup.find('span')
        if url:
            name = url['title']
            url = url['href']
            assert isinstance(url, str)
            bangumi_id = url.split('/')[-1]
            soup.find('li', )
            li.append({'name': name, 'keyword': bangumi_id, 'cover': span['data-src']})
    return li


def parser_subtitle_of_bangumi(bangumi_id):
    bangumi_id = int(bangumi_id)
    url = "https://mikanani.me/Home/ExpandBangumi"
    data = {'bangumiId': bangumi_id, 'showSubscribed': False}
    if os.environ.get('DEBUG', False):
        print(url, data)
    r = requests.post(url, data=data, ).text
    soup = BeautifulSoup(r, 'lxml')
    g = soup.find('ul', class_='list-unstyled res-ul')
    subtitle_list = []
    for li in g.find_all('li', class_='js-expand_bangumi-subgroup', ):
        subgroup_index = li['data-bangumisubgroupindex']
        name = li.find('div', class_='sk-col tag-res-name').text
        _id = li.find('div', class_='btn-primary ladda-button sk-col tag-sub js-subscribe_bangumi')[
            'data-subtitlegroupid']
        id_div = li.find('div', class_=['js-subscribe_bangumi'])
        subgroup_div = soup.find('div', class_='js-expand_bangumi-subgroup-{}-episodes'.format(subgroup_index))
        ul_e = subgroup_div.find('ul', class_='list-unstyled res-detail-ul')
        episode_list = []
        for li2 in ul_e.find_all('li'):
            magnet_url = li2.find('a', class_='js-magnet magnet-link')
            episode_title = li2.find('a', class_='magnet-link-wrap')
            if magnet_url:
                episode_list.append({
                    '_id': _id,
                    'title': episode_title.text,
                    'url': magnet_url['data-clipboard-text'],
                    # 'url': 'mag',
                    'publish_time': li2.find('div', class_='sk-col res-date').text
                })
        subtitle_list.append({
            'id': id_div['data-subtitlegroupid'],
            'name': name,
            'episode': [] if not episode_list else episode_list
        })
    return subtitle_list


class Mikanani(BaseWebsite):
    cover_url = 'https://mikanani.me'

    def search_by_keyword(self, keyword, count):
        """
        return a list of dict with at least 4 key: download, name, title, episode
        example:
        ```
            [
                {
                    'name':"路人女主的养成方法",
                    'download': 'magnet:?xt=urn:btih:what ever',
                    'title': "[澄空学园] 路人女主的养成方法 第12话 MP4 720p  完",
                    'episode': 12
                },
            ]
        ```
        :param keyword: search key word
        :type keyword: str
        :param count: how many page to fetch from website
        :type count: int

        :return: list of episode search result
        :rtype: list[dict]
        """

        result = []
        if os.environ.get('DEBUG', False):
            print('https://mikanani.me/Home/Search', {'searchstr': keyword})
        r = requests.get('https://mikanani.me/Home/Search', params={'searchstr': keyword}).text
        s = BeautifulSoup(r, 'lxml')
        td_list = s.find_all('tr', attrs={'class': 'js-search-results-row'})  # type:list[bs4.Tag]
        for tr in td_list:
            title = tr.find('a', class_='magnet-link-wrap').text
            time_string = tr.find_all('td')[2].string
            result.append({
                'download': tr.find('a', class_='magnet-link').attrs.get('data-clipboard-text', ''),
                'name': keyword,
                'title': title,
                'episode': self.parse_episode(title),
                'time': int(time.mktime(time.strptime(time_string, "%Y/%m/%d %H:%M")))
            })
            # print(result)
        return result

    def fetch_episode_of_bangumi(self, bangumi_id, subtitle_list=None, max_page=MAX_PAGE):
        """
        get all episode by bangumi id
        example
        ```
            [
                {
                    "download": "magnet:?xt=urn:btih:e43b3b6b53dd9fd6af1199e112d3c7ff15cab82c",
                    "name": "来自深渊",
                    "subtitle_group": "58a9c1c9f5dc363606ab42ec",
                    "title": "【喵萌奶茶屋】★七月新番★[来自深渊/Made in Abyss][07][GB][720P]",
                    "episode": 0,
                    "time": 1503301292
                },
            ]
        ```

        :param bangumi_id: bangumi_id
        :param subtitle_list: list of subtitle group
        :type subtitle_list: list
        :param max_page: how many page you want to crawl if there is no subtitle list
        :type max_page: int
        :return: list of bangumi
        :rtype: list[dict]
        """

        result = []
        if os.environ.get('DEBUG', False):
            print('https://mikanani.me/Home/Bangumi/{}'.format(bangumi_id))
        r = requests.get('https://mikanani.me/Home/Bangumi/{}'.format(bangumi_id)).text

        soup = BeautifulSoup(r, 'lxml')
        # name = soup.find('p', class_='bangumi-title').text
        container = soup.find('div', class_='central-container')  # type:bs4.Tag
        episode_container_list = {}
        for index, tag in enumerate(container.contents):
            if hasattr(tag, 'attrs'):
                subtitle_id = tag.attrs.get('id', False)
                if subtitle_list:
                    if subtitle_id in subtitle_list:
                        episode_container_list[tag.attrs.get('id', None)] = tag.find_next_sibling('table')
                else:
                    if subtitle_id:
                        episode_container_list[tag.attrs.get('id', None)] = tag.find_next_sibling('table')

        for subtitle_id, container in episode_container_list.items():
            for tr in container.find_all('tr')[1:]:
                title = tr.find('a', class_='magnet-link-wrap').text
                time_string = tr.find_all('td')[2].string
                result.append({
                    'download': tr.find('a', class_='magnet-link').attrs.get('data-clipboard-text', ''),
                    'subtitle_group': str(subtitle_id),
                    'title': title,
                    'episode': self.parse_episode(title),
                    'time': int(time.mktime(time.strptime(time_string, "%Y/%m/%d %H:%M")))
                })

        return result

    def fetch_bangumi_calendar_and_subtitle_group(self):
        """
        return a list of all bangumi and a list of all subtitle group

        list of bangumi dict:
        update time should be one of ['Sun', 'Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat']
        example:
        ```
            [
                {
                    "status": 0,
                    "subtitle_group": [
                        "123",
                        "456"
                    ],
                    "name": "名侦探柯南",
                    "keyword": "1234", #bangumi id
                    "update_time": "Sat",
                    "cover": "data/images/cover1.jpg"
                },
            ]
        ```

        list of subtitle group dict:
        example:
        ```
            [
                {
                    'id': '233',
                    'name': 'bgmi字幕组'
                }
            ]
        ```


        :return: list of bangumi, list of subtitile group
        :rtype: (list[dict], list[dict])
        """
        bangumi_result = []
        subtitle_result = []
        for index, day in enumerate(get_weekly_bangumi()):
            for obj in parser_day_bangumi(day):
                obj['update_time'] = week[index]
                bangumi_result.append(obj)
        for index, bangumi in enumerate(bangumi_result):
            # time.sleep(0.1)
            # print('fetching ', bangumi['name'])
            subtitle_list = parser_subtitle_of_bangumi(bangumi_id=bangumi['keyword'])
            bangumi_result[index]['subtitle_group'] = list(map(lambda x: x['id'], subtitle_list))
            bangumi_result[index]['name'] = bangumi['name']
            bangumi_result[index]['status'] = 0
            for subtitle in subtitle_list:
                subtitle_result.append({'id': subtitle['id'], 'name': subtitle['name']})
        # print('fetching end')
        return bangumi_result, subtitle_result



# coding=utf-8
import os

from bgmi.config import BGMI_SAVE_PATH, BGMI_PATH, DOWNLOAD_DELEGATE, BGMI_TMP_PATH, SCRIPT_PATH
from bgmi.download import get_download_class
from bgmi.utils import print_success, print_warning, print_info, print_error


def install_crontab():
    print_info('Installing crontab job')
    path = os.path.join(os.path.dirname(__file__), 'crontab.sh')
    os.system('sh \'%s\'' % path)


def create_dir():
    if not os.environ.get('HOME', ''):
        print_warning('$HOME not set, use \'/tmp/\'')

    tools_path = os.path.join(BGMI_PATH, 'tools')
    # bgmi home dir
    path_to_create = (BGMI_PATH, BGMI_SAVE_PATH, BGMI_TMP_PATH, SCRIPT_PATH, tools_path)

    try:
        for path in path_to_create:
            if not os.path.exists(path):
                print_success('%s created successfully' % path)
                os.mkdir(path)
            else:
                print_warning('%s already exists' % path)
    except OSError as e:
        print_error('Error: {0}'.format(str(e)))


def install():
    get_download_class(DOWNLOAD_DELEGATE, instance=False).install()



# coding=utf-8
import os
import sqlite3

from bgmi import __version__
from bgmi.config import DB_PATH, BGMI_PATH, SCRIPT_DB_PATH
from bgmi.utils import print_error, print_info

OLD = os.path.join(BGMI_PATH, 'old')


def exec_sql(sql, db=DB_PATH):
    try:
        print_info('Execute {}'.format(sql))
        conn = sqlite3.connect(db)
        conn.execute(sql)
        conn.commit()
        conn.close()
    except sqlite3.OperationalError:
        print_error('Execute SQL statement failed', exit_=False)


def update_database():
    if not os.path.exists(OLD):
        v = '0'
        with open(OLD, 'w') as f:
            f.write(__version__)
    else:
        with open(OLD, 'r+b') as f:
            v = f.read()
            f.seek(0)
            f.write(__version__)

    if v < '1.0.25':
        exec_sql('ALTER TABLE filter ADD COLUMN regex')

    if v < '1.4.1':
        exec_sql('ALTER TABLE scripts ADD COLUMN update_time INTEGER', SCRIPT_DB_PATH)



# coding=utf-8
from __future__ import print_function, unicode_literals

import argparse
import codecs
import locale
import os
import platform
import signal
import sqlite3
import sys

import bgmi.config
from bgmi.config import BGMI_PATH, DB_PATH, SCRIPT_DB_PATH
from bgmi.constants import *
from bgmi.controllers import controllers
from bgmi.sql import (CREATE_TABLE_BANGUMI, CREATE_TABLE_FOLLOWED, CREATE_TABLE_DOWNLOAD, CREATE_TABLE_FOLLOWED_FILTER,
                      CREATE_TABLE_SUBTITLE, CREATE_TABLE_SCRIPT)
from bgmi.update import update_database
from bgmi.utils import print_warning, print_error, print_version, check_update

# Wrap sys.stdout into a StreamWriter to allow writing unicode.
if bgmi.config.IS_PYTHON3:
    unicode = str
    if platform.system() != 'Windows':
        file_ = sys.stdout.buffer
        sys.stdout = codecs.getwriter(locale.getpreferredencoding())(file_)
else:
    sys.stdout = codecs.getwriter('utf-8')(sys.stdout)
    input = raw_input


# global Ctrl-C signal handler
def signal_handler(signal, frame):
    print_error('User aborted, quit')


signal.signal(signal.SIGINT, signal_handler)


def unicode_(s):
    if not bgmi.config.IS_PYTHON3:
        unicode_string = s.decode(sys.getfilesystemencoding())
        return unicode_string
    else:
        return unicode(s)


# main function
def main():
    c = argparse.ArgumentParser()

    c.add_argument('--version', help='Show the version of BGmi.', action='version', version=print_version())

    sub_parser = c.add_subparsers(help='BGmi actions', dest='action')
    sub_parser_add = sub_parser.add_parser(ACTION_ADD, help='Subscribe bangumi.')
    sub_parser_add.add_argument('name', metavar='name', type=unicode_, nargs='+', help='Bangumi name')
    sub_parser_add.add_argument('--episode', metavar='episode', help='Add bangumi and mark it as specified episode.',
                                type=int)
    sub_parser_add.add_argument('--not-ignore', action='store_true',
                                help='Do not ignore the old bangumi detail rows (3 month ago).')

    sub_parser_list = sub_parser.add_parser(ACTION_LIST, help='List subscribed bangumi.')

    sub_parser_filter = sub_parser.add_parser(ACTION_FILTER, help='Set bangumi fetch filter.')
    sub_parser_filter.add_argument('name', metavar='name', type=unicode_, help='Bangumi name to set the filter.')
    sub_parser_filter.add_argument('--subtitle', metavar='subtitle', type=unicode_,
                                   help='Subtitle group name, split by ",".')
    sub_parser_filter.add_argument('--include', metavar='include', type=unicode_,
                                   help='Filter by keywords which in the title, split by ",".')
    sub_parser_filter.add_argument('--exclude', metavar='exclude', type=unicode_,
                                   help='Filter by keywords which not int the title, split by ",".')
    sub_parser_filter.add_argument('--regex', metavar='regex', type=unicode_,
                                   help='Filter by regular expression')

    sub_parser_del = sub_parser.add_parser(ACTION_DELETE, help='Unsubscribe bangumi.')
    sub_parser_del_mutex = sub_parser_del.add_mutually_exclusive_group(required=True)
    sub_parser_del_mutex.add_argument('--name', metavar='name', nargs='+', type=unicode_,
                                      help='Bangumi name to unsubscribe.')
    sub_parser_del_mutex.add_argument('--clear-all', action='store_true',
                                      help='Clear all the subscriptions.')
    sub_parser_del.add_argument('--batch', action='store_true', help='No confirmation.')

    sub_parser_update = sub_parser.add_parser(ACTION_UPDATE, help='Update bangumi calendar and '
                                                                  'subscribed bangumi episode.')
    sub_parser_update.add_argument('name', metavar='name', type=unicode_, nargs='*', help='Update specified bangumi.')
    sub_parser_update.add_argument('--download', action='store',
                                   help='Download specified episode of the bangumi when updated.',
                                   nargs='*', type=int, metavar='episode')
    sub_parser_update.add_argument('--not-ignore', action='store_true',
                                   help='Do not ignore the old bangumi detail rows (3 month ago).')

    sub_parser_cal = sub_parser.add_parser(ACTION_CAL, help='Print bangumi calendar.')
    sub_parser_cal.add_argument('--today', action='store_true', help='Show bangumi calendar for today.')
    sub_parser_cal.add_argument('--force-update', action='store_true',
                                help='Get the newest bangumi calendar from bangumi.moe.')
    sub_parser_cal.add_argument('--no-save', action='store_true',
                                help='Do not save the bangumi data when force update.')

    sub_parser_config = sub_parser.add_parser(ACTION_CONFIG, help='Config BGmi.')
    sub_parser_config.add_argument('name', nargs='?', type=unicode_, help='Config name')
    sub_parser_config.add_argument('value', nargs='?', type=unicode_, help='Config value')

    sub_parser_mark = sub_parser.add_parser(ACTION_MARK, help='Mark bangumi episode.')
    sub_parser_mark.add_argument('name', type=unicode_, help='Bangumi name')
    sub_parser_mark.add_argument('episode', help='Bangumi episode', type=int)

    sub_parser_download = sub_parser.add_parser(ACTION_DOWNLOAD, help='Download manager.')
    sub_parser_download.add_argument('--list', help='List download queue.', action='store_true')
    sub_parser_download.add_argument('--mark', help='Mark download status with a specific id.', dest='id', type=int)
    sub_parser_download.add_argument('--status', type=int, help='Download items status (0: not download, 1: '
                                                                'downloading, 2: alrealy downloaded).',
                                     choices=[0, 1, 2])

    sub_parser_fetch = sub_parser.add_parser(ACTION_FETCH, help='Fetch bangumi.')
    sub_parser_fetch.add_argument('name', help='Bangumi name', type=unicode_)
    sub_parser_fetch.add_argument('--not-ignore', action='store_true',
                                  help='Do not ignore the old bangumi detail rows (3 month ago).')

    sub_parser_search = sub_parser.add_parser(ACTION_SEARCH)
    sub_parser_search.add_argument('keyword', help='Search keyword', type=unicode_)
    sub_parser_search.add_argument('--count', type=int, help='The max page count of search result.')
    sub_parser_search.add_argument('--regex-filter', type=unicode_, help='Regular expression filter of title.')
    sub_parser_search.add_argument('--download', action='store_true',
                                   help='Download search result.')

    sub_parser_search = sub_parser.add_parser(ACTION_SOURCE, help='select date source bangumi_moe or mikan_project')
    sub_parser_search.add_argument('source', help='bangumi_moe or mikan_project', type=unicode_)


    sub_parser.add_parser('install', help='Install BGmi download delegate.')
    sub_parser.add_parser('upgrade', help='Check update.')

    ret = c.parse_args()

    if ret.action == 'install':
        import bgmi.setup

        bgmi.setup.install()
        raise SystemExit
    elif ret.action == 'upgrade':
        update_database()
        check_update(mark=False)
    else:
        check_update()
        controllers(ret)


def init_db():
    try:
        # bangumi.db
        conn = sqlite3.connect(DB_PATH)
        conn.execute(CREATE_TABLE_BANGUMI)
        conn.execute(CREATE_TABLE_FOLLOWED)
        conn.execute(CREATE_TABLE_DOWNLOAD)
        conn.execute(CREATE_TABLE_FOLLOWED_FILTER)
        conn.execute(CREATE_TABLE_SUBTITLE)
        conn.commit()
        conn.close()

        # script.db
        conn = sqlite3.connect(SCRIPT_DB_PATH)
        conn.execute(CREATE_TABLE_SCRIPT)
        conn.commit()
        conn.close()
    except sqlite3.OperationalError:
        print_error('Open database file failed, path %s is not writable.' % BGMI_PATH)


def setup():
    if not os.path.exists(BGMI_PATH):
        print_warning('BGMI_PATH %s does not exist, installing' % BGMI_PATH)
        from bgmi.setup import create_dir, install_crontab

        create_dir()
        if not platform.system() == 'Windows':
            # if not input('Do you want to install a crontab to auto-download bangumi?(Y/n): ') == 'n':
            install_crontab()

    # if not os.path.exists(DB_PATH):
    init_db()
    main()


if __name__ == '__main__':
    setup()




# coding=utf-8
from __future__ import print_function, unicode_literals

import time

from bgmi.config import SUPPORT_WEBSITE
from bgmi.config import write_config
from bgmi.constants import *
from bgmi.download import download_prepare
from bgmi.download import get_download_class
from bgmi.fetch import website
from bgmi.models import (Bangumi, Followed, Download, Filter, Subtitle, STATUS_FOLLOWED, STATUS_UPDATED,
                         STATUS_NORMAL, STATUS_NOT_DOWNLOAD, DB)
from bgmi.script import ScriptRunner
from bgmi.utils import print_warning, print_info, print_success, print_error


def add(ret):
    # action add
    # add bangumi by a list of bangumi name
    if not Bangumi.get_all_bangumi():
        print_warning('No bangumi data in database, fetching...')
        website.fetch(save=True, group_by_weekday=False)

    for bangumi in ret.name:
        bangumi_obj = Bangumi(name=bangumi)
        data = bangumi_obj.select(one=True, fields=['id', 'name', 'keyword'])
        if data:
            followed_obj = Followed(bangumi_name=data['name'], status=STATUS_FOLLOWED)
            followed_obj.select_obj()
            if not followed_obj or followed_obj.status == STATUS_NORMAL:
                if not followed_obj:
                    bangumi_data, _ = website.get_maximum_episode(bangumi_obj, subtitle=False, max_page=1)
                    followed_obj.episode = bangumi_data['episode'] if ret.episode is None else ret.episode
                    followed_obj.save()
                else:
                    followed_obj.status = STATUS_FOLLOWED
                    followed_obj.save()
                print_success('{0} has been followed'.format(bangumi_obj))
            else:
                print_warning('{0} already followed'.format(bangumi_obj))

        else:
            print_error('{0} not found, please check the name'.format(bangumi))


def print_filter(followed_filter_obj):
    print_info('Followed subtitle group: {0}'.format(', '.join(map(lambda s: s['name'],
                                                                   Subtitle.get_subtitle(
                                                                       followed_filter_obj.subtitle.split(', '))))
                                                     if followed_filter_obj.subtitle else 'None'))
    print_info('Include keywords: {0}'.format(followed_filter_obj.include))
    print_info('Exclude keywords: {0}'.format(followed_filter_obj.exclude))
    print_info('Regular expression: {0}'.format(followed_filter_obj.regex))


def filter_(ret):
    bangumi_obj = Bangumi(name=ret.name)
    bangumi_obj.select_obj()
    if not bangumi_obj:
        print_error('Bangumi {0} does not exist.'.format(bangumi_obj.name))

    followed_obj = Followed(bangumi_name=bangumi_obj.name)
    followed_obj.select_obj()

    if not followed_obj:
        print_error('Bangumi {0} has not subscribed, try \'bgmi add "{1}"\'.'.format(bangumi_obj.name,
                                                                                     bangumi_obj.name))

    subtitle = ret.subtitle
    include = ret.include
    exclude = ret.exclude
    regex = ret.regex

    followed_filter_obj = Filter(bangumi_name=ret.name)
    followed_filter_obj.select_obj()

    if not followed_filter_obj:
        followed_filter_obj.save()

    if subtitle is not None:
        subtitle = map(lambda s: s.strip(), subtitle.split(','))

        subtitle = map(lambda s: s['id'], Subtitle.get_subtitle_by_name(subtitle))

        subtitle_list = [s.split('.')[0] for s in bangumi_obj.subtitle_group.split(', ')
                         if '.' in s]
        subtitle_list.extend(bangumi_obj.subtitle_group.split(', '))
        subtitle = filter(lambda s: True if s in subtitle_list else False, subtitle)
        subtitle = ', '.join(subtitle)
        followed_filter_obj.subtitle = subtitle

    if include is not None:
        followed_filter_obj.include = include

    if exclude is not None:
        followed_filter_obj.exclude = exclude

    if regex is not None:
        followed_filter_obj.regex = regex

    followed_filter_obj.save()
    print_info('Usable subtitle group: {0}'.format(', '.join(map(lambda s: s['name'],
                                                                 Subtitle.get_subtitle(
                                                                     bangumi_obj.subtitle_group.split(', ')))))
               if bangumi_obj.subtitle_group else 'None')

    print_filter(followed_filter_obj)


def delete(ret):
    # action delete
    # just delete subscribed bangumi or clear all the subscribed bangumi
    if ret.clear_all:
        if Followed.delete_followed(batch=ret.batch):
            print_success('all subscriptions have been deleted')
        else:
            print_error('user canceled')
    elif ret.name:
        for name in ret.name:
            followed = Followed(bangumi_name=name)
            if followed.select():
                followed.delete()
                print_warning('Bangumi %s has been deleted' % name)
            else:
                print_error('Bangumi %s does not exist' % name, exit_=False)
    else:
        print_warning('Nothing has been done.')


def update(ret):
    ignore = not bool(ret.not_ignore)
    print_info('marking bangumi status ...')
    now = int(time.time())
    for i in Followed.get_all_followed():
        if i['updated_time'] and int(i['updated_time'] + 86400) < now:
            followed_obj = Followed(bangumi_name=i['bangumi_name'])
            followed_obj.status = STATUS_FOLLOWED
            followed_obj.save()

    for script in ScriptRunner().scripts:
        obj = script.Model().obj
        if obj['updated_time'] and int(obj['updated_time'] + 86400) < now:
            obj.status = STATUS_FOLLOWED
            obj.save()

    print_info('updating bangumi data ...')
    website.fetch(save=True, group_by_weekday=False)
    print_info('updating subscriptions ...')
    download_queue = []

    if ret.download:
        if not ret.name:
            print_warning('No specified bangumi, ignore `--download` option')
        if len(ret.name) > 1:
            print_warning('Multiple specified bangumi, ignore `--download` option')

    if not ret.name:
        updated_bangumi_obj = Followed.get_all_followed()
    else:
        updated_bangumi_obj = []
        for i in ret.name:
            f = Followed(bangumi_name=i)
            f.select_obj()
            updated_bangumi_obj.append(f)

    runner = ScriptRunner()
    script_download_queue = runner.run()

    for subscribe in updated_bangumi_obj:
        print_info('fetching %s ...' % subscribe['bangumi_name'])
        bangumi_obj = Bangumi(name=subscribe['bangumi_name'])
        bangumi_obj.select_obj()

        followed_obj = Followed(bangumi_name=subscribe['bangumi_name'])
        followed_obj.select_obj()

        # filter by subtitle group
        if not bangumi_obj or not followed_obj:
            print_error('Bangumi<{0}> does not exist or not been followed.'.format(subscribe['bangumi_name']),
                        exit_=False)
            continue

        episode, all_episode_data = website.get_maximum_episode(bangumi=bangumi_obj, ignore_old_row=ignore, max_page=1)

        if (episode.get('episode') > subscribe['episode']) or (len(ret.name) == 1 and ret.download):
            if len(ret.name) == 1 and ret.download:
                episode_range = ret.download
            else:
                episode_range = range(subscribe['episode'] + 1, episode.get('episode', 0) + 1)
                print_success('%s updated, episode: %d' % (subscribe['bangumi_name'], episode['episode']))
                followed_obj.episode = episode['episode']
                followed_obj.status = STATUS_UPDATED
                followed_obj.updated_time = int(time.time())
                followed_obj.save()

            for i in episode_range:
                for epi in all_episode_data:
                    if epi['episode'] == i:
                        download_queue.append(epi)
                        break

    if ret.download is not None:
        download_prepare(download_queue)
        download_prepare(script_download_queue)
        print_info('Re-downloading ...')
        download_prepare(Download.get_all_downloads(status=STATUS_NOT_DOWNLOAD))


def cal(ret):
    website.bangumi_calendar(force_update=ret.force_update, today=ret.today, save=not ret.no_save)


def download_manager(ret):
    if ret.id:
        download_id = ret.id
        status = ret.status
        if download_id is None or status is None:
            print_error('No id or status specified.')
        download_obj = Download(_id=download_id)
        download_obj.select_obj()
        if not download_obj:
            print_error('Download object does not exist.')
        print_info('Download Object <{0} - {1}>, Status: {2}'.format(download_obj.name, download_obj.episode,
                                                                     download_obj.status))
        download_obj.status = status
        download_obj.save()
        print_success('Download status has been marked as {0}'.format(DOWNLOAD_CHOICE_LIST_DICT.get(int(status))))
    else:
        status = ret.status
        status = int(status) if status is not None else None
        delegate = get_download_class(instance=False)
        delegate.download_status(status=status)


def mark(ret):
    name = ret.name
    episode = ret.episode
    followed_obj = Followed(bangumi_name=name)
    followed_obj.select_obj()

    runner = ScriptRunner()

    if not followed_obj:
        followed_obj = runner.get_model(name)

    if not followed_obj:
        print_error('Subscribe or Script <%s> does not exist.' % name)

    if episode is not None:
        followed_obj.episode = episode
        followed_obj.save()
        print_success('%s has been mark as episode: %s' % (followed_obj, followed_obj.episode))
    else:
        print_info('%s, episode: %s' % (followed_obj, followed_obj.episode))


def followed(ret):
    if ret.list:
        website.bangumi_calendar(followed=True, save=False)
    else:
        mark(ret)


def list_(ret):
    website.bangumi_calendar(followed=True, save=False)


def fetch_(ret):
    bangumi_obj = Bangumi(name=ret.name)
    bangumi_obj.select_obj()

    followed_obj = Followed(bangumi_name=bangumi_obj.name)
    followed_obj.select_obj()

    followed_filter_obj = Filter(bangumi_name=ret.name)
    followed_filter_obj.select_obj()
    print_filter(followed_filter_obj)

    if bangumi_obj:
        print_info('Fetch bangumi {0} ...'.format(bangumi_obj.name))
        _, data = website.get_maximum_episode(bangumi_obj,
                                              ignore_old_row=False if ret.not_ignore else True)
        if not data:
            print_warning('Nothing.')
        for i in data:
            print_success(i['title'])

    else:
        print_error('Bangumi {0} not exist'.format(ret.name))


def search_(ret):
    if not ret.count:
        ret.count = 3

    data = website.search(ret.keyword, count=ret.count, filter_=ret.regex_filter)

    for i in data:
        print_success(i['title'])
    if ret.download:
        download_prepare(data)


def source(ret):
    if ret.source in list(map(lambda x: x['id'], SUPPORT_WEBSITE)):
        print_success('you select {}'.format(ret.source))
        DB.recreate_source_relatively_table()
        write_config('DATA_SOURCE', ret.source)
        print_success('data source switch succeeds')
        from bgmi.fetch import DATA_SOURCE_MAP

        DATA_SOURCE_MAP.get(ret.source)().bangumi_calendar(force_update=True)
    else:
        print_error(
            'please check input.nata source should be {} or {}'.format(*list(map(lambda x: x['id'], SUPPORT_WEBSITE))))


def search_without_filter(keyword):
    data = website.raw_search(keyword, count=3)
    return data


def config(ret):
    write_config(ret.name, ret.value)


CONTROLLERS_DICT = {
    ACTION_ADD: add,
    ACTION_FILTER: filter_,
    ACTION_CAL: cal,
    ACTION_DELETE: delete,
    ACTION_DOWNLOAD: download_manager,
    ACTION_UPDATE: update,
    ACTION_FETCH: fetch_,
    ACTION_CONFIG: config,
    ACTION_FOLLOWED: followed,
    ACTION_MARK: mark,
    ACTION_LIST: list_,
    ACTION_SEARCH: search_,
    ACTION_SOURCE: source,
}


def controllers(ret):
    func = CONTROLLERS_DICT.get(ret.action, None)
    if func is None or not callable(func):
        return
    else:
        return func(ret)



# coding=utf-8
from __future__ import print_function, unicode_literals

import datetime
import os
import time
import requests

from bgmi.config import (LANG, MAX_PAGE, COVER_URL, BANGUMI_MOE_URL)
from bgmi.utils import (print_info, print_error, bug_report)
from bgmi.models import Bangumi

from bgmi.website.base import BaseWebsite

# tag of bangumi on bangumi.moe
BANGUMI_TAG = '549ef207fe682f7549f1ea90'

__split = '/' if not BANGUMI_MOE_URL.endswith('/') else ''
FETCH_URL = '{0}{1}api/bangumi/current'.format(BANGUMI_MOE_URL, __split)
TEAM_URL = '{0}{1}api/team/working'.format(BANGUMI_MOE_URL, __split)
NAME_URL = '{0}{1}api/tag/fetch'.format(BANGUMI_MOE_URL, __split)
DETAIL_URL = '{0}{1}api/torrent/search'.format(BANGUMI_MOE_URL, __split)
SEARCH_URL = '{0}{1}api/v2/torrent/search'.format(BANGUMI_MOE_URL, __split)


def get_response(url, method='GET', **kwargs):
    if os.environ.get('DEBUG'):
        print_info('Request URL: {0}'.format(url))
    try:
        if os.environ.get('DEBUG'):
            print(getattr(requests, method.lower())(url, **kwargs).text)
        return getattr(requests, method.lower(), )(url, **kwargs).json()
    except requests.ConnectionError:
        print_error('error: failed to establish a new connection')
    except ValueError:
        print_error('error: server returned data maybe not be json, please contact ricterzheng@gmail.com')


def process_name(data):
    lang = 'zh_cn' if LANG not in ('zh_cn', 'zh_tw', 'ja', 'en') else LANG
    return {i['_id']: i['locale'][lang] for i in data}


def process_subtitle(data):
    """get subtitle group name from links
    """
    result = {}
    for s in data:
        # pprint(data)
        # f = Subtitle(id=s['tag_id'], name=s['name'])
        # if not f.select():
        #     f.save()
        result[s['tag_id']] = s['name']
    return result


def parser_bangumi(data):
    """match weekly bangumi list from data
    """

    ids = list(map(lambda b: b['tag_id'], data))
    subtitle = get_response(TEAM_URL, 'POST', json={'tag_ids': ids})
    name = process_name(get_response(NAME_URL, 'POST', json={'_ids': ids}))

    weekly_list = []
    subtitle_group_list = []
    for bangumi_item in data:
        subtitle_of_bangumi = process_subtitle(subtitle.get(bangumi_item['tag_id'], []))
        item = {'status': 0,
                'subtitle_group': subtitle_of_bangumi.keys(),
                'name': name[bangumi_item['tag_id']],
                'keyword': bangumi_item['tag_id'],
                'update_time': Bangumi.week[bangumi_item['showOn'] - 1],
                'cover': bangumi_item['cover']}
        for key, value in subtitle_of_bangumi.items():
            subtitle_group_list.append({
                'id': key,
                'name': value
            })

        weekly_list.append(item)

    if not weekly_list:
        bug_report()
    return weekly_list, subtitle_group_list


class BangumiMoe(BaseWebsite):
    cover_url = COVER_URL

    def fetch_episode_of_bangumi(self, bangumi_id, subtitle_list=None, max_page=MAX_PAGE):
        """
        get all episode by bangumi id
        example
        ```
            [
                {
                    "download": "magnet:?xt=urn:btih:e43b3b6b53dd9fd6af1199e112d3c7ff15cab82c",
                    "name": "来自深渊",
                    "subtitle_group": "58a9c1c9f5dc363606ab42ec",
                    "title": "【喵萌奶茶屋】★七月新番★[来自深渊/Made in Abyss][07][GB][720P]",
                    "episode": 0,
                    "time": 1503301292
                },
            ]
        ```

        :param bangumi_id: bangumi_id
        :param subtitle_list: list of subtitle group
        :type subtitle_list: list
        :param max_page: how many page you want to crawl if there is no subtitle list
        :type max_page: int
        :return: list of bangumi
        :rtype: list[dict]
        """

        response_data = []
        if subtitle_list:
            for subtitle_id in subtitle_list:
                data = {'tag_id': [bangumi_id, subtitle_id, BANGUMI_TAG]}
                response = get_response(DETAIL_URL, 'POST', json=data)
                response_data.extend(response['torrents'])
        else:
            for i in range(max_page):
                if max_page > 1:
                    print_info('Fetch page {0} ...'.format(i + 1))
                data = {'tag_id': [bangumi_id, BANGUMI_TAG], 'p': i + 1}
                response = get_response(DETAIL_URL, 'POST', json=data)
                if response:
                    response_data.extend(response['torrents'])
        for index, bangumi in enumerate(response_data):
            response_data[index] = {
                'download': bangumi['magnet'],
                'subtitle_group': bangumi['team_id'],
                'title': bangumi['title'],
                'episode': self.parse_episode(bangumi['title']),
                'time': int(time.mktime(datetime.datetime.strptime(bangumi['publish_time'].split('.')[0],
                                                                   "%Y-%m-%dT%H:%M:%S").timetuple()))
            }

        return response_data

    def fetch_bangumi_calendar_and_subtitle_group(self):
        """
        return a list of all bangumi and a list of all subtitle group

        list of bangumi dict:
        update time should be one of ['Sun', 'Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat']
        example:
        ```
            [
                {
                    "status": 0,
                    "subtitle_group": [
                        "123",
                        "456"
                    ],
                    "name": "名侦探柯南",
                    "keyword": "1234", #bangumi id
                    "update_time": "Sat",
                    "cover": "data/images/cover1.jpg"
                },
            ]
        ```

        list of subtitle group dict:
        example:
        ```
            [
                {
                    'id': '233',
                    'name': 'bgmi字幕组'
                }
            ]
        ```


        :return: list of bangumi, list of subtitile group
        :rtype: (list[dict], list[dict])
        """
        response = get_response(FETCH_URL)
        if not response:
            return []

        bangumi_result, subtitile_result = parser_bangumi(response)

        return bangumi_result, subtitile_result

    def search_by_keyword(self, keyword, count):
        """
        return a list of dict with at least 4 key: download, name, title, episode
        example:
        ```
            [
                {
                    'name':"路人女主的养成方法",
                    'download': 'magnet:?xt=urn:btih:what ever',
                    'title': "[澄空学园] 路人女主的养成方法 第12话 MP4 720p  完",
                    'episode': 12
                },
            ]

        :param keyword: search key word
        :type keyword: str
        :param count: how many page to fetch from website
        :type count: int

        :return: list of episode search result
        :rtype: list[dict]
        """

        rows = []
        result = []

        for i in range(count):
            data = get_response(SEARCH_URL, 'POST', json={'query': keyword, 'p': i + 1})
            rows.extend(data['torrents'])

        for info in rows:
            if True:
                result.append({
                    'download': info['magnet'],
                    'name': keyword,
                    'subtitle_group': info['team_id'],
                    'title': info['title'],
                    'episode': self.parse_episode(info['title']),
                    'time': int(time.mktime(datetime.datetime.strptime(info['publish_time'].split('.')[0],
                                                                       "%Y-%m-%dT%H:%M:%S").timetuple()))
                })

        # Avoid bangumi collection. It's ok but it will waste your traffic and bandwidth.
        result = result[::-1]
        return result



# coding=utf-8
from __future__ import print_function, unicode_literals
import os
import sys
import time
import platform
import struct
import functools
import requests
from bgmi import __version__
from bgmi.config import IS_PYTHON3, BGMI_PATH, DATA_SOURCE, SUPPORT_WEBSITE

requests.packages.urllib3.disable_warnings()


if platform.system() == 'Windows':
    GREEN = ''
    YELLOW = ''
    RED = ''
    COLOR_END = ''
else:
    GREEN = '\033[1;32m'
    YELLOW = '\033[1;33m'
    RED = '\033[1;31m'
    COLOR_END = '\033[0m'

color_map = {
    'print_info': '',
    'print_success': GREEN,
    'print_warning': YELLOW,
    'print_error': RED,
}

indicator_map = {
    'print_info': '[*] ',
    'print_success': '[+] ',
    'print_warning': '[-] ',
    'print_error': '[x] ',
}


def indicator(f):
    @functools.wraps(f)
    def wrapper(*args, **kwargs):
        if kwargs.get('indicator', True):
            if IS_PYTHON3:
                func_name = f.__qualname__
            else:
                func_name = f.func_name
            args = (indicator_map.get(func_name, '') + args[0],)
        f(*args, **kwargs)
        sys.stdout.flush()

    return wrapper


def colorize(f):
    @functools.wraps(f)
    def wrapper(*args, **kwargs):
        if IS_PYTHON3:
            func_name = f.__qualname__
        else:
            func_name = f.func_name
        b, e = color_map.get(func_name, ''), COLOR_END if color_map.get(func_name) else ''
        args = tuple(map(lambda s: b + s + e, args))
        return f(*args, **kwargs)

    return wrapper


@indicator
@colorize
def print_info(message, indicator=True):
    print(message)


@indicator
@colorize
def print_success(message, indicator=True):
    print(message)


@indicator
@colorize
def print_warning(message, indicator=True):
    print(message)


@indicator
@colorize
def print_error(message, exit_=True, indicator=True):
    print(message)
    if exit_:
        exit(1)


def print_version():
    return '''BGmi %sver. %s%s built by %sRicterZ%s with ❤️

Github: https://github.com/RicterZ/BGmi
Email: ricterzheng@gmail.com
Blog: https://ricterz.me''' % (YELLOW, __version__, COLOR_END, YELLOW, COLOR_END)


def test_connection():
    try:
        for website in SUPPORT_WEBSITE:
            if DATA_SOURCE == website['id']:
                requests.head(website['url'], timeout=10)
    except:
        return False

    return True


def unicodeize(data):
    import bgmi.config

    if bgmi.config.IS_PYTHON3:
        if isinstance(data, bytes):
            return data.decode('utf-8')
        else:
            return data
            # return bytes(str(data), 'latin-1').decode('utf-8')
    try:
        return unicode(data.decode('utf-8'))
    except (UnicodeEncodeError, UnicodeDecodeError):
        return unicode(data.decode('gbk'))
    except (UnicodeEncodeError, UnicodeDecodeError):
        return data


def bug_report():
    print_error('It seems that no bangumi found, if https://bangumi.moe can \n'
                '    be opened normally, please report bug to ricterzheng@gmail.com\n'
                '    or submit issue at: https://github.com/RicterZ/BGmi/issues',
                exit_=True)


def get_terminal_col():
    # https://gist.github.com/jtriley/1108174
    if not platform.system() == 'Windows':
        import fcntl
        import termios

        _, col, _, _ = struct.unpack(str('HHHH'), fcntl.ioctl(0, termios.TIOCGWINSZ,
                                                              struct.pack(str('HHHH'), 0, 0, 0, 0)))

        return col
    else:
        try:
            from ctypes import windll, create_string_buffer

            # stdin handle is -10
            # stdout handle is -11
            # stderr handle is -12
            h = windll.kernel32.GetStdHandle(-12)
            csbi = create_string_buffer(22)
            res = windll.kernel32.GetConsoleScreenBufferInfo(h, csbi)
            if res:
                (bufx, bufy, curx, cury, wattr,
                 left, top, right, bottom,
                 maxx, maxy) = struct.unpack("hhhhHhhhhhh", csbi.raw)
                sizex = right - left + 1
                # sizey = bottom - top + 1
                return sizex
            else:
                import subprocess

                cols = int(subprocess.check_output('tput cols'))
                return cols
        except:
            return 80


def check_update(mark=True):
    def update():
        print_info('Checking update ...')
        version = requests.get('https://pypi.python.org/pypi/bgmi/json', verify=False).json()['info']['version']
        if version > __version__:
            print_warning('Please update bgmi to the latest version {}{}{}.'
                          '\nThen execute `bgmi upgrade` to migrate database'.format(GREEN, version, COLOR_END))
        else:
            print_success('Your BGmi is the latest version.')

    if not mark:
        update()
        raise SystemExit

    version_file = os.path.join(BGMI_PATH, 'version')
    if not os.path.exists(version_file):
        with open(version_file, 'w') as f:
            f.write(str(int(time.time())))
        return update()

    with open(version_file, 'r') as f:
        try:
            data = int(f.read())
            if time.time() - 7 * 24 * 3600 > data:
                with open(version_file, 'w') as f:
                    f.write(str(int(time.time())))
                return update()
        except ValueError:
            pass



# coding=utf-8
from __future__ import print_function, unicode_literals

import os

from bgmi.config import BGMI_SAVE_PATH, DOWNLOAD_DELEGATE
from bgmi.models import Download, STATUS_DOWNLOADING, STATUS_NOT_DOWNLOAD
from bgmi.services import XunleiLixianDownload, Aria2DownloadRPC, RRDownload, TransmissionRPC
from bgmi.utils import print_error

DOWNLOAD_DELEGATE_DICT = {
    'xunlei': XunleiLixianDownload,
    'aria2-rpc': Aria2DownloadRPC,
    'transmission-rpc': TransmissionRPC,
    'rr!': RRDownload,
}


def get_download_class(download_obj=None, save_path='', overwrite=True, instance=True):
    if DOWNLOAD_DELEGATE not in DOWNLOAD_DELEGATE_DICT:
        print_error('unexpected delegate {0}'.format(DOWNLOAD_DELEGATE))

    delegate = DOWNLOAD_DELEGATE_DICT.get(DOWNLOAD_DELEGATE)

    if instance:
        delegate = delegate(download_obj=download_obj, overwrite=overwrite, save_path=save_path)

    return delegate


def download_prepare(data):
    queue = save_to_bangumi_download_queue(data)
    for download in queue:
        save_path = os.path.join(os.path.join(BGMI_SAVE_PATH, download.name), str(download.episode))
        # mark as downloading
        download.status = STATUS_DOWNLOADING
        download.save()
        try:
            # start download
            download_class = get_download_class(download_obj=download, save_path=save_path)
            download_class.download()
            download_class.check_download(download.name)

            # mark as downloaded
            download.delete()
        except Exception as e:
            print_error('Error: {0}'.format(e), exit_=False)
            download.status = STATUS_NOT_DOWNLOAD
            download.save()


def save_to_bangumi_download_queue(data):
    queue = []
    for i in data:
        download = Download(status=STATUS_NOT_DOWNLOAD, name=i['name'], title=i['title'],
                            episode=i['episode'], download=i['download'])
        download.save()
        queue.append(download)

    return queue



# coding=utf-8
from __future__ import print_function, unicode_literals

import os
import sqlite3
import sys
from collections import defaultdict

import bgmi.config
from bgmi.config import IS_PYTHON3
from bgmi.sql import (CLEAR_TABLE_)

if IS_PYTHON3:
    _unicode = str
else:
    input = raw_input
    _unicode = unicode


# bangumi subscription and download status
STATUS_UPDATING = 0
STATUS_END = 1
BANGUMI_STATUS = (STATUS_UPDATING, STATUS_END)

STATUS_NORMAL = 0
STATUS_FOLLOWED = 1
STATUS_UPDATED = 2
FOLLOWED_STATUS = (STATUS_NORMAL, STATUS_FOLLOWED, STATUS_UPDATED)

STATUS_NOT_DOWNLOAD = 0
STATUS_DOWNLOADING = 1
STATUS_DOWNLOADED = 2
DOWNLOAD_STATUS = (STATUS_NOT_DOWNLOAD, STATUS_DOWNLOADING, STATUS_DOWNLOADED)


def make_dicts(cursor, row):
    return dict((cursor.description[idx][0], value)
                for idx, value in enumerate(row))


class FalseType(type):
    def __nonzero__(self):
        return False


class DB(object):
    db_path = ''

    # `_id` save the id column in database, will be set automatic
    _id = None

    # `primary_key` is one of the fields in a table, which maybe `UNIQUE` or `PRIMARY_KEY`.
    # It will be set when instantiate a DB object, and `primary_key` is used as query condition
    # when `_id` is `None`.
    primary_key = ()

    # all columns of a table except `id`, must be a sequence instance.
    fields = ()

    # table name
    table = None
    _conn = None

    # nonzero
    __nonzero = True

    def __init__(self, **kwargs):
        if '_id' in kwargs:
            self._id = kwargs.get('_id')

        for f in self.fields:
            if f in self.primary_key and kwargs.get(f, None) is None:
                if '_id' not in kwargs:
                    raise ValueError('primary key %s must be set' % f)
            setattr(self, f, kwargs.get(f, None))

        self._unicodeize()
        self.select(one=True)

    def __getitem__(self, item):
        return getattr(self, item)

    @staticmethod
    def recreate_source_relatively_table():
        db = DB.connect_db()
        table_to_drop = ['bangumi', 'followed', 'subtitle', 'filter', 'download']
        for table in table_to_drop:
            sql = CLEAR_TABLE_.format(table)
            db.execute(sql)

        cur = db.cursor()
        cur.close()
        DB.close_db(db)
        return True

    @staticmethod
    def _make_sql(method, table, fields=None, data=None, condition=None, operation='AND',
                  join=None, order=None, desc=None):
        '''
        Make SQL statement (just a simple implementation, don't support complex operation).

        :param method: expect `select`, `update`, `delete`, `insert`
        :param table: the main table name of the SQL statement
        :param fields: fields will be operated
        :param data:
        :param condition: conditions, only support sequences
        :param join:
        :return:
        '''
        if method not in ('select', 'update', 'delete', 'insert'):
            raise Exception('unexpected operation %s' % method)

        if not isinstance(condition, (type(None), tuple, list, set, _unicode)):
            raise Exception('`condition` expected sequences')

        if not isinstance(fields, (type(None), tuple, list, set, _unicode)):
            raise Exception('`select` expected sequences or string')

        if not isinstance(table, _unicode):
            raise Exception('unexpected type %s of table' % type(table))

        def make_condition(condition, operation='AND'):
            if not condition:
                return '1'

            sql = ''

            if isinstance(condition, _unicode):
                condition = (condition, )

            if isinstance(condition, (tuple, list, set)):
                for f in condition:
                    if f.startswith('!'):
                        operator = '!'
                        f = f[1:]
                    else:
                        operator = ''

                    if '.' in f:
                        name = '%s' % f
                    else:
                        name = '`%s`' % f
                    sql += '%s%s=? %s ' % (name, operator, operation)

                sql = sql[:-(len(operation)+1)]
            else:
                sql += '1'
            return sql

        def make_fields(fields):
            sql = ''
            for f in fields:
                if '.' in f:
                    sql += '%s,' % f
                else:
                    sql += '`%s`,' % f
            sql = sql[:-1]
            return sql

        if method == 'insert':
            sql = 'INSERT INTO %s ' % table
            if fields is not None:
                sql += '(%s)' % make_fields(fields)

            sql += ' VALUES ('
            for i in range(len(fields)):
                sql += '?,'
            sql = sql[:-1]
            sql += ')'

        elif method == 'select':

            if fields is None:
                select = '*'
            else:
                if not isinstance(fields, _unicode):
                    select = ''
                    for f in fields:
                        if '.' in f:
                            select += '%s,' % f
                        else:
                            select += '`%s`,' % f
                    select = select[:-1]
                else:
                    select = '`%s`' % fields

            if not isinstance(join, _unicode):
                join = ''

            sql = 'SELECT %s FROM `%s` %s WHERE ' % (select, table, join)
            sql += make_condition(condition, operation=operation)
            if order:
                if '.' in order:
                    sql += 'ORDER BY %s ' % order
                else:
                    sql += 'ORDER BY `%s` ' % order
            if desc:
                sql += 'DESC'

        elif method == 'update':
            sql = 'UPDATE %s SET ' % table
            if fields is not None:
                sql += make_condition(fields, ',')
            else:
                raise Exception('UPDATE: unexpected null fields')

            sql += ' WHERE '
            sql += make_condition(condition)
        elif method == 'delete':
            sql = 'DELETE FROM %s WHERE ' % table
            if condition is not None:
                sql += make_condition(condition)
            else:
                sql += '1'

        if os.environ.get('DEBUG'):
            # print('EXEC SQL: {0}'.format(sql))
            pass

        return sql

    def _unicodeize(self):
        for i in self.fields:
            v = self.__dict__.get(i, '')
            if isinstance(v, str):
                if sys.version_info.major < 3:
                    v = unicode(v.decode('utf-8'))
                else:
                    v = str(v)
                self.__dict__[i] = v

    @staticmethod
    def connect_db():
        return sqlite3.connect(bgmi.config.DB_PATH)

    @staticmethod
    def close_db(db_instance):
        db_instance.commit()
        db_instance.close()

    def _connect_db(self):
        self._conn = sqlite3.connect(bgmi.config.DB_PATH if not self.db_path else self.db_path)
        self._conn.row_factory = make_dicts
        self.cursor = self._conn.cursor()

    def _close_db(self):
        self._conn.commit()
        self.cursor.close()
        self._conn.close()

    def _pair(self):
        values = tuple([self.__dict__.get(i, '') for i in self.fields])
        return self.fields, values

    def select_obj(self):
        data = self.select(one=True)
        if not data:
            self.__nonzero = False
        else:
            self.__nonzero = True
            for k, v in data.items():
                setattr(self, k, v)

    def __bool__(self):
        return self.__nonzero

    def __nonzero__(self):
        return self.__nonzero

    def select(self, fields=None, condition=None, one=False, join=None):
        if not isinstance(condition, (dict, type(None))):
            raise Exception('condition expected dict')

        if condition is None:
            if self._id is None:
                if self.primary_key:
                    k, v = self.primary_key, [self.__dict__.get(i, '') for i in self.primary_key]
                else:
                    k = []
                    v = []
                    for i in self.fields:
                        if self.__dict__.get(i, None):
                            k.append(i)
                            v.append(self.__dict__.get(i))
            else:
                k, v = '%s.id' % self.table, (self._id, )
        else:
            # hack for python3
            k, v = list(condition.keys()), list(condition.values())

        self._connect_db()
        sql = Bangumi._make_sql('select', fields=fields, table=self.table, condition=k, join=join)
        self.cursor.execute(sql, v)
        ret = self.cursor.fetchall() if not one else self.cursor.fetchone()
        self._close_db()

        if self._id is None and ret and one and 'id' in ret:
            self._id = ret['id']

        if not isinstance(ret, (list, type(None))):
            for i in self.fields:
                if getattr(self, i) is None:
                    if i in ret:
                        setattr(self, i, ret[i])

        return ret

    def update(self, data=None):
        obj = self.select(one=True)
        if obj:
            self._id = obj['id']
        else:
            raise Exception('%s not exist' % self.__repr__())

        if not isinstance(data, (dict, type(None))):
            raise Exception('update data expected dict')

        if data is None:
            data = {}
            for i in self.fields:
                data.update({i: getattr(self, i)})

        sql = self._make_sql('update', self.table, fields=list(data.keys()), condition=('id', ))
        self._connect_db()
        params = list(data.values())
        params.append(self._id)
        self.cursor.execute(sql, params)
        self._close_db()

    def delete(self, condition=None):
        if not self._id:
            obj = self.select(one=True)
            if not obj:
                raise Exception('%s does not exist' % self.__repr__())

        sql = self._make_sql('delete', self.table, condition=('id', ))
        self._connect_db()
        self.cursor.execute(sql, (self._id, ))
        self._close_db()

    def save(self):
        _f, _v = self._pair()

        obj = self.select(one=True)
        if obj:
            self._id = obj['id']
            self.update()
            return self

        self._connect_db()
        sql = self._make_sql('insert', self.table, _f)
        self.cursor.execute(sql, _v)
        self._id = self.cursor.lastrowid
        self._close_db()

        return self

    @staticmethod
    def execute(sql):
        db = DB.connect_db()
        db.execute(sql)
        DB.close_db(db)


class Bangumi(DB):
    table = 'bangumi'
    primary_key = ('name', )
    fields = ('name', 'update_time', 'subtitle_group', 'keyword', 'status', 'cover', )
    week = ('Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat', 'Sun')

    def __init__(self, **kwargs):
        super(Bangumi, self).__init__(**kwargs)

        update_time = kwargs.get('update_time', '').title()
        if update_time and update_time not in self.week:
            raise ValueError('unexcept update time %s' % update_time)
        self.update_time = update_time
        self.subtitle_group = ', '.join(kwargs.get('subtitle_group', []))
        self._unicodeize()
        self.select(one=True)

    def __repr__(self):
        return 'Bangumi<%s>' % self.name

    def __str__(self):
        return 'Bangumi<%s>' % self.name

    @staticmethod
    def delete_all():
        db = Bangumi.connect_db()
        sql = Bangumi._make_sql('update', table=Bangumi.table, fields=('status'))
        cur = db.cursor()
        cur.execute(sql, (STATUS_END, ))
        Bangumi.close_db(db)

    @staticmethod
    def get_all_bangumi(status=None, order=True):
        db = Bangumi.connect_db()
        db.row_factory = make_dicts
        cur = db.cursor()
        join_sql = Bangumi._make_sql('select', table=Followed.table)
        if status is None:
            sql = Bangumi._make_sql('select', fields=['%s.*' % Bangumi.table, 'F.status AS status',
                                                      'episode'], table=Bangumi.table,
                                    condition=('%s.status' % Bangumi.table),
                                    join='LEFT JOIN (%s) AS F ON bangumi.name=F.bangumi_name' % join_sql)
            cur.execute(sql, (STATUS_UPDATING, ))
        else:
            sql = Bangumi._make_sql('select', fields=['%s.*' % Bangumi.table, 'F.status AS status',
                                                      'episode'], table=Bangumi.table,
                                    join='LEFT JOIN (%s) AS F ON bangumi.name=F.bangumi_name' % join_sql,
                                    condition=('F.status', '%s.status' % Bangumi.table))
            cur.execute(sql, (status, STATUS_UPDATING, ))
        data = cur.fetchall()
        Bangumi.close_db(db)

        if order:
            weekly_list = defaultdict(list)
            for bangumi_item in data:
                weekly_list[bangumi_item['update_time'].lower()].append(dict(bangumi_item))
        else:
            weekly_list = data

        return weekly_list


class Followed(DB):
    table = 'followed'
    primary_key = ('bangumi_name', )
    fields = ('bangumi_name', 'episode', 'status', 'updated_time',)

    @staticmethod
    def delete_followed(condition=None, batch=True):
        db = DB.connect_db()
        db.row_factory = make_dicts
        cur = db.cursor()
        if not isinstance(condition, (type(None), dict)):
            raise Exception('condition expected dict')
        if condition is None:
            k, v = [], []
        else:
            k, v = list(condition.keys()), list(condition.values())
        sql = DB._make_sql('delete', table=Followed.table, condition=k)

        if not batch and sql.endswith('WHERE 1'):
            if not input('[+] are you sure want to CLEAR ALL THE BANGUMI? (y/N): ') == 'y':
                return False

        cur.execute(sql, v)
        DB.close_db(db)
        return True

    def delete(self, condition=None):
        self.status = STATUS_NORMAL
        self.save()

    @staticmethod
    def get_all_followed(status=STATUS_NORMAL, bangumi_status=STATUS_UPDATING, order=None, desc=None):
        db = DB.connect_db()
        db.row_factory = make_dicts
        cur = db.cursor()
        if status is None and bangumi_status is None:
            sql = DB._make_sql('select', fields=['followed.*', 'bangumi.cover', 'bangumi.update_time'],
                               table=Followed.table,
                               join='LEFT JOIN bangumi on bangumi.name=followed.bangumi_name', order=order,
                               desc=desc)
            cur.execute(sql)
        else:
            sql = DB._make_sql('select', fields=['followed.*', 'bangumi.cover', 'bangumi.update_time'],
                               table=Followed.table,
                               join='LEFT JOIN bangumi on bangumi.name=followed.bangumi_name',
                               condition=['!followed.status', 'bangumi.status'], order=order, desc=desc)
            cur.execute(sql, (status, bangumi_status))
        data = cur.fetchall()
        DB.close_db(db)
        return data

    def __str__(self):
        return 'Followed Bangumi<%s>' % self.bangumi_name

    def __repr__(self):
        return 'Followed Bangumi<%s>' % self.bangumi_name


class Download(DB):
    table = 'download'
    primary_key = ('name', 'episode', )
    fields = ('name', 'title', 'episode', 'download', 'status', )

    @staticmethod
    def get_all_downloads(status=None):
        db = DB.connect_db()
        db.row_factory = make_dicts
        cur = db.cursor()

        if status is None:
            sql = DB._make_sql('select', table=Download.table)
            sql += ' order by status'
            cur.execute(sql)
        else:
            sql = DB._make_sql('select', table=Download.table, condition=['status', ])
            cur.execute(sql, (status, ))

        data = cur.fetchall()
        DB.close_db(db)
        return data

    def delete(self, condition=None):
        self.status = STATUS_DOWNLOADED
        self.save()


class Filter(DB):
    table = 'filter'
    primary_key = ('bangumi_name', )
    fields = ('bangumi_name', 'subtitle', 'include', 'exclude', 'regex', )


class Subtitle(DB):
    table = 'subtitle'
    primary_key = ('id', )
    fields = ('id', 'name', )

    @staticmethod
    def get_subtitle(l=None):
        l = list(l)
        db = DB.connect_db()
        db.row_factory = make_dicts
        cur = db.cursor()
        sql = DB._make_sql('select', fields='name', table=Subtitle.table,
                           condition=['id'] * len(l), operation='OR')
        cur.execute(sql, l)
        data = cur.fetchall()
        DB.close_db(db)
        return data

    @staticmethod
    def get_subtitle_by_name(l=None):
        l = list(l)
        db = DB.connect_db()
        db.row_factory = make_dicts
        cur = db.cursor()
        sql = DB._make_sql('select', fields='id', table=Subtitle.table,
                           condition=['name'] * len(l), operation='OR')
        cur.execute(sql, l)
        data = cur.fetchall()
        DB.close_db(db)
        return data


class Script(DB):
    db_path = bgmi.config.SCRIPT_DB_PATH
    table = 'scripts'
    primary_key = ('bangumi_name', )
    fields = ('bangumi_name', 'episode', 'status', 'updated_time',)

    def __str__(self):
        return 'Script Model <%s>' % self.bangumi_name


# coding=utf-8
from __future__ import print_function, unicode_literals

from bgmi.config import DATA_SOURCE
from bgmi.utils import print_error
from bgmi.website import bangumi_moe, mikan

DATA_SOURCE_MAP = {
    'mikan_project': mikan.Mikanani,
    'bangumi_moe': bangumi_moe.BangumiMoe
}


def wrap(*args,**kwargs):
    print_error('date source "{}" in config is wrong, please edit it manually'.format(DATA_SOURCE))

    pass


website = DATA_SOURCE_MAP.get(DATA_SOURCE, wrap)()



# coding=utf-8
from __future__ import print_function, unicode_literals

import re
import os
import imp
import glob
import time
import traceback

from bgmi.models import STATUS_UPDATED, STATUS_FOLLOWED
from bgmi.utils import print_success, print_warning, print_info
from bgmi.config import SCRIPT_PATH
from bgmi.download import download_prepare
from bgmi.models import Script


FETCH_EPISODE_ZH = re.compile("第?\s?(\d{1,3})\s?[話话集]")
FETCH_EPISODE_WITH_BRACKETS = re.compile('[【\[](\d+)\s?(?:END)?[】\]]')
FETCH_EPISODE_ONLY_NUM = re.compile('^([\d]{2,})$')
FETCH_EPISODE_RANGE = re.compile('[\d]{2,}\s?-\s?([\d]{2,})')
FETCH_EPISODE_OVA_OAD = re.compile('([\d]{2,})\s?\((?:OVA|OAD)\)]')
FETCH_EPISODE_WITH_VERSION = re.compile('[【\[](\d+)\s? *v\d(?:END)?[】\]]')
FETCH_EPISODE = (
    FETCH_EPISODE_ZH, FETCH_EPISODE_WITH_BRACKETS, FETCH_EPISODE_ONLY_NUM,
    FETCH_EPISODE_RANGE,
    FETCH_EPISODE_OVA_OAD, FETCH_EPISODE_WITH_VERSION)


def parse_episode(episode_title):
    """
    parse episode from title

    :param episode_title: episode title
    :type episode_title: str
    :return: episode of this title
    :rtype: int
    """

    _ = FETCH_EPISODE_ZH.findall(episode_title)
    if _ and _[0].isdigit():
        return int(_[0])

    _ = FETCH_EPISODE_WITH_BRACKETS.findall(episode_title)
    if _ and _[0].isdigit():
        return int(_[0])

    _ = FETCH_EPISODE_WITH_VERSION.findall(episode_title)
    if _ and _[0].isdigit():
        return int(_[0])

    for split_token in ['【', '[', ' ']:
        for i in episode_title.split(split_token):
            for regexp in FETCH_EPISODE:
                match = regexp.findall(i)
                if match and match[0].isdigit():
                    return int(match[0])
    return 0


class ScriptRunner(object):
    _defined = None
    scripts = []
    download_queue = []

    def __new__(cls, *args, **kwargs):
        if cls._defined is None:

            script_files = glob.glob('{}{}*.py'.format(SCRIPT_PATH, os.path.sep))
            for i in script_files:
                try:
                    s = imp.load_source('script', os.path.join(SCRIPT_PATH, i))
                    script_class = getattr(s, 'Script')()
                    cls.scripts.append(script_class)
                    print_success('Load script {} successfully.'.format(i))
                except:
                    print_warning('Load script {} failed, ignored'.format(i))
                    traceback.print_exc()
                    # self.scripts = filter(self._check_followed, self.scripts)
                    # self.scripts = filter(self._check_bangumi, self.scripts)

            cls._defined = super(ScriptRunner, cls).__new__(cls, *args, **kwargs)

        return cls._defined

    def get_model(self, name):
        for script in self.scripts:
            if script.Model.bangumi_name == name:
                return script.Model().obj

    def get_models_dict(self):
        return [dict(script.Model()) for script in self.scripts if script.bangumi_name is not None]

    @staticmethod
    def make_dict(script):
        return [{
            'name': script.bangumi_name,
            'title': '[{}][{}]'.format(script.bangumi_name, k),
            'episode': k,
            'download': v
        } for k, v in script.get_download_url().items()]

    def run(self, return_=True, download=False):
        for script in self.scripts:
            print_info('fetching {} ...'.format(script.bangumi_name))
            download_item = self.make_dict(script)

            script_obj = script.Model().obj

            if not download_item:
                print_info('Got nothing, quit script {}.'.format(script))
                break

            max_episode = max(download_item, key=lambda d: d['episode'])
            episode = max_episode['episode']
            episode_range = range(script_obj.episode + 1, episode + 1)

            if episode <= script_obj.episode:
                break

            print_success('{} updated, episode: {}'.format(script.bangumi_name, episode))
            script_obj.episode = episode
            script_obj.status = STATUS_UPDATED
            script_obj.updated_time = int(time.time())
            script_obj.save()

            download_queue = []
            for i in episode_range:
                for e in download_item:
                    if i == e['episode']:
                        download_queue.append(e)

            if return_:
                self.download_queue.extend(download_queue)
                continue

            if download:
                print_success('Start downloading of {}'.format(script))
                download_prepare(download_queue)

        return self.download_queue


class ScriptBase(object):

    class Model(object):
        obj = None
        bangumi_name = None
        cover = None
        update_time = None

        def __init__(self):
            if self.bangumi_name is not None:
                s = Script(bangumi_name=self.bangumi_name, episode=0, status=STATUS_FOLLOWED)
                s.select_obj()
                if not s:
                    s.save()
                self.obj = s

        def __iter__(self):
            for i in ('bangumi_name', 'cover', 'update_time'):
                yield (i, getattr(self, i))

            # patch for cal
            yield ('update_time', self.update_time)
            yield ('name', self.bangumi_name)
            yield ('status', self.obj['status'])
            yield ('subtitle_group', '')
            yield ('episode', self.obj['episode'])

    @property
    def name(self):
        return self.Model.bangumi_name

    @property
    def bangumi_name(self):
        return self.Model.bangumi_name

    @property
    def cover(self):
        return self.Model.cover

    @property
    def updated_time(self):
        return self.Model.update_time

    def __unicode__(self):
        return self.__str__()

    def __str__(self):
        return '<Script of \'{}\'>'.format(self.bangumi_name)

    def get_download_url(self):
        """Get the download url, and return a dict of episode and the url.
        Download url also can be magnet link.
        For example:
        ```
            {
                1: 'http://example.com/Bangumi/1/1.mp4'
                2: 'http://example.com/Bangumi/1/2.mp4'
                3: 'http://example.com/Bangumi/1/3.mp4'
            }
        ```
        The keys `1`, `2`, `3` is the episode, the value is the url of bangumi.
        :return: dict
        """
        return {}


if __name__ == '__main__':
    runner = ScriptRunner()
    runner.run()



# coding=utf-8
import os
import codecs
from setuptools import setup, find_packages
from bgmi import __version__, __author__, __email__


with open('requirements.txt', 'r') as f:
    requirements = f.read().splitlines()


ROOT = os.path.abspath(os.path.dirname(__file__))


def long_description():
    with codecs.open('README.rst', 'rb') as f:
        return f.read().decode('utf-8')


setup(
    name='bgmi',
    version=__version__,
    author=__author__,
    author_email=__email__,
    keywords='bangumi, bgmi, feed',
    description='BGmi is a cli tool for subscribed bangumi.',
    long_description=long_description(),
    url='https://github.com/RicterZ/BGmi',
    download_url='https://github.com/RicterZ/BGmi/tarball/master',
    packages=find_packages(),
    package_data={'': ['LICENSES']},
    include_package_data=True,
    zip_safe=False,
    install_requires=requirements,
    entry_points={
        'console_scripts': [
            'bgmi = bgmi.main:setup',
            'bgmi_http = bgmi.front.http:main'
        ]
    },
    license='MIT License',
    classifiers=(
        'Development Status :: 4 - Beta',
        'Intended Audience :: Other Audience',
        'Natural Language :: Chinese (Traditional)',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy'
    ),
)



# coding=utf-8
from __future__ import unicode_literals, print_function

import os
import subprocess
from tempfile import NamedTemporaryFile

import bgmi.config
from bgmi.config import XUNLEI_LX_PATH, BGMI_PATH, BGMI_TMP_PATH, ARIA2_RPC_URL, ARIA2_RPC_TOKEN, \
    WGET_PATH, TRANSMISSION_RPC_PORT, TRANSMISSION_RPC_URL

from bgmi.utils import print_warning, print_info, print_error, print_success
from bgmi.models import Download, STATUS_DOWNLOADED, STATUS_NOT_DOWNLOAD, STATUS_DOWNLOADING


if bgmi.config.IS_PYTHON3:
    from xmlrpc.client import ServerProxy, _Method
else:
    from xmlrpclib import ServerProxy, _Method


class _PatchedMethod(_Method):
    def __getitem__(self, name):
        return _Method.__getattr__(self, name)

    def __getattr__(self, name):
        if name == '__getitem__':
            return self.__getitem__
        return _Method.__getattr__(self, name)

    def __call__(self, *args):
        print(args)


class PatchedServerProxy(ServerProxy):
    def __request(self, methodname, params):
        return ServerProxy._ServerProxy__request(self, methodname, params)

    def __getattr__(self, name):
        return _PatchedMethod(self.__request, name)


#######################
#   DownloadService   #
#######################
class DownloadService(object):
    def __init__(self, download_obj, save_path, overwrite=True):
        self.name = download_obj.name
        self.torrent = download_obj.download
        self.overwrite = overwrite
        self.save_path = save_path
        self.episode = download_obj.episode
        self.return_code = 0

    def download(self):
        # download
        raise NotImplementedError

    @staticmethod
    def install():
        # install requirement
        raise NotImplementedError

    def check_path(self):
        if not os.path.exists(self.save_path):
            print_warning('Create dir {0}'.format(self.save_path))
            os.makedirs(self.save_path)

    def check_delegate_bin_exist(self, path):
        if not os.path.exists(path):
            raise Exception('{0} not exist, please run command \'bgmi install\' to install'.format(path))

    def call(self, command):
        self.return_code = subprocess.call(command, env={'PATH': '/usr/local/bin:/usr/bin:/bin',
                                                         'HOME': os.environ.get('HOME', '/tmp')})

    def check_download(self, name):
        if not os.path.exists(self.save_path) or self.return_code != 0:
            raise Exception('It seems the bangumi {0} not be downloaded'.format(name))

    @staticmethod
    def download_status(status=None):
        last_status = -1
        for download_data in Download.get_all_downloads(status=status):
            latest_status = download_data['status']
            name = '  {0}. <{1}: {2}>'.format(download_data['id'], download_data['name'],
                                              download_data['episode'])
            if latest_status != last_status:
                if latest_status == STATUS_DOWNLOADING:
                    print('Downloading items:')
                elif latest_status == STATUS_NOT_DOWNLOAD:
                    print('Not downloaded items:')
                elif latest_status == STATUS_DOWNLOADED:
                    print('Downloaded items:')

            if download_data['status'] == STATUS_NOT_DOWNLOAD:
                print_info(name, indicator=False)
            elif download_data['status'] == STATUS_DOWNLOADING:
                print_warning(name, indicator=False)
            elif download_data['status'] == STATUS_DOWNLOADED:
                print_success(name, indicator=False)
            last_status = download_data['status']


class TransmissionRPC(DownloadService):

    def __init__(self, *args, **kwargs):
        self.check_delegate_bin_exist('')
        super(TransmissionRPC, self).__init__(*args, **kwargs)

    def download(self):
        try:
            import transmissionrpc
            tc = transmissionrpc.Client(TRANSMISSION_RPC_URL, port=TRANSMISSION_RPC_PORT)
            tc.add_torrent(self.torrent, download_dir=self.save_path)
            print_info('Add torrent into the download queue, the file will be saved at {0}'.format(self.save_path))
        except ImportError:
            pass

    def check_delegate_bin_exist(self, path):
        try:
            import transmissionrpc
        except ImportError:
            raise Exception('Please run `pip install transmissionrpc`')

    def check_download(self, name):
        pass

    @staticmethod
    def download_status(status=None):
        print_info('Print download status in database')
        DownloadService.download_status(status=status)
        print()
        print_info('Print download status in transmission-rpc')
        try:
            import transmissionrpc
            tc = transmissionrpc.Client(TRANSMISSION_RPC_URL, port=TRANSMISSION_RPC_PORT)
            for torrent in tc.get_torrents():
                print_info('  * {0}: {1}'.format(torrent.status, torrent), indicator=False)
        except ImportError:
            pass

    @staticmethod
    def install():
        print_warning('Please run `pip install transmissionrpc`')


class Aria2DownloadRPC(DownloadService):
    old_version = False

    def __init__(self, *args, **kwargs):
        self.server = PatchedServerProxy(ARIA2_RPC_URL)
        Aria2DownloadRPC.check_aria2c_version()
        super(Aria2DownloadRPC, self).__init__(**kwargs)

    def download(self):
        if self.old_version:
            self.server.aria2.addUri([self.torrent], {"dir": self.save_path})
        else:
            self.server.aria2.addUri(ARIA2_RPC_TOKEN, [self.torrent], {"dir": self.save_path})
        print_info('Add torrent into the download queue, the file will be saved at {0}'.format(self.save_path))

    @staticmethod
    def check_aria2c_version():
        url = ARIA2_RPC_URL.split('/')
        url[2] = ARIA2_RPC_TOKEN + '@' + url[2]
        url = '/'.join(url)
        s = ServerProxy(url)
        r = s.aria2.getVersion(ARIA2_RPC_TOKEN, )
        version = r['version']
        if version:
            Aria2DownloadRPC.old_version = version < '1.18.4'
        else:
            print_warning('Get aria2c version failed')

    @staticmethod
    def install():
        print_warning('Please install aria2 by yourself')

    def check_download(self, name):
        pass

    @staticmethod
    def download_status(status=None):
        Aria2DownloadRPC.check_aria2c_version()

        print_info('Print download status in database')
        DownloadService.download_status(status=status)
        print()
        print_info('Print download status in aria2c-rpc')
        try:
            server = PatchedServerProxy(ARIA2_RPC_URL)
            # self.server.aria2
            status_dict = {
                STATUS_DOWNLOADING: ['tellActive'],
                STATUS_NOT_DOWNLOAD: ['tellWaiting'],
                STATUS_DOWNLOADED: ['tellStopped'],
                None: ['tellStopped', 'tellWaiting', 'tellActive'],
            }
            for method in status_dict.get(status):
                if method not in ('tellActive', ):
                    params = (0, 1000)
                else:
                    params = ()
                if Aria2DownloadRPC.old_version:
                    data = server.aria2[method](*params)
                else:
                    data = server.aria2[method](ARIA2_RPC_TOKEN, *params)

                if data:
                    print_warning('RPC {0}:'.format(method), indicator=False)

                for row in data:
                    print_success('- {0}'.format(row['dir']), indicator=False)
                    for file_ in row['files']:
                        print_info('    * {0}'.format(file_['path']), indicator=False)

        except Exception as e:
            print_error('Cannot connect to aria2-rpc server')


class XunleiLixianDownload(DownloadService):
    def __init__(self, *args, **kwargs):
        self.check_delegate_bin_exist(XUNLEI_LX_PATH)
        super(XunleiLixianDownload, self).__init__(*args, **kwargs)

    def download(self):
        overwrite = '--overwrite' if self.overwrite else ''

        command = [XUNLEI_LX_PATH, 'download', '--torrent', overwrite,
                   '--output-dir={0}'.format(self.save_path), self.torrent,
                   '--verification-code-path={0}'.format(os.path.join(BGMI_TMP_PATH, 'vcode.jpg'))]

        print_info('Run command {0}'.format(' '.join(command)))
        print_warning('Verification code path: {0}'.format(os.path.join(BGMI_TMP_PATH, 'vcode.jpg')))
        self.call(command)

    @staticmethod
    def install():
        # install xunlei-lixian
        import tarfile
        import requests
        print_info('Downloading xunlei-lixian from https://github.com/iambus/xunlei-lixian/')
        r = requests.get('https://github.com/iambus/xunlei-lixian/tarball/master', stream=True,
                         headers={'Accept-Encoding': ''})
        f = NamedTemporaryFile(delete=False)

        with f:
            for chunk in r.iter_content(chunk_size=1024):
                if chunk:
                    f.write(chunk)
        f.close()
        print_success('Download successfully, save at %s, extracting ...' % f.name)
        zip_file = tarfile.open(f.name, 'r:gz')
        zip_file.extractall(os.path.join(BGMI_PATH, 'tools/xunlei-lixian'))
        dir_name = zip_file.getnames()[0]

        print_info('Create link file ...')

        if not os.path.exists(XUNLEI_LX_PATH):
            os.symlink(os.path.join(BGMI_PATH, 'tools/xunlei-lixian/{0}/lixian_cli.py'.format(dir_name)),
                       XUNLEI_LX_PATH)
        else:
            print_warning('{0} already exists'.format(XUNLEI_LX_PATH))

        print_success('All done')
        print_info('Please run command \'{0} config\' to configure your lixian-xunlei '
                   '(Notice: only for Thunder VIP)'.format(XUNLEI_LX_PATH))


class RRDownload(DownloadService):
    def __init__(self, *args, **kwargs):
        self.check_delegate_bin_exist(WGET_PATH)
        super(RRDownload, self).__init__(*args, **kwargs)

    @staticmethod
    def install():
        print_info('Please install curl by yourself')

    def download(self):
        if not os.path.exists(self.save_path):
            os.makedirs(self.save_path)

        command = [WGET_PATH, '--no-parent', '-r', '--no-host-directories',
                   '--cut-dirs', '100', '--reject', 'index.html',
                   '-P', '{0}/'.format(self.save_path),
                   'https://bgmi.ricterz.me/bangumi/{0}/{1}/'.format(self.name, self.episode)]
        print_info('Start download ...')
        self.call(command)


#######################
#   SendMailService   #
#######################
class SendMailService(object):
    pass



# coding=utf-8
from __future__ import print_function, unicode_literals
import os
import unittest
import sqlite3
from bgmi.models import Bangumi, Followed, STATUS_FOLLOWED, STATUS_NORMAL
from bgmi.sql import CREATE_TABLE_BANGUMI, CREATE_TABLE_FOLLOWED, CREATE_TABLE_DOWNLOAD
import bgmi.config

bgmi.config.DB_PATH = '/tmp/bangumi.db'
DB_PATH = bgmi.config.DB_PATH


class ModelsTest(unittest.TestCase):
    def setUp(self):
        if not os.path.exists(DB_PATH):
            self.db = sqlite3.connect(DB_PATH)
            self.conn = self.db.cursor()
            self.conn.execute(CREATE_TABLE_BANGUMI)
            self.conn.execute(CREATE_TABLE_FOLLOWED)
            self.conn.execute(CREATE_TABLE_DOWNLOAD)
        else:
            self.db = sqlite3.connect(DB_PATH)
            self.conn = self.db.cursor()
        self.db.commit()

    def tearDown(self):
        self.db.close()
        os.remove(DB_PATH)

    def test_select_and_save(self):
        b2 = Bangumi(name='test_select_and_save', update_time='Sun')
        d = b2.select()
        self.assertEqual(d, [])
        b2.save()
        ret = b2.select(one=True)
        self.assertIsInstance(ret, dict)

    def test_get_all_bangumi(self):
        from collections import defaultdict
        self.assertIsInstance(Bangumi.get_all_bangumi(), defaultdict)

    def test_update(self):
        b1 = Bangumi(name='test_update', update_time='Sun')
        b1.save()
        b1.update_time = 'Mon'
        b1.update()
        ret = b1.select(one=True)
        self.assertEqual(ret['update_time'], 'Mon')

        b1.update(data={'name': 'test666', 'update_time': 'Sat'})
        ret = b1.select(one=True)
        self.assertIsInstance(ret, dict)

    def test_delete(self):
        b1 = Bangumi(name='test_delete', update_time='Sun')
        b1.save()
        self.assertEqual(b1.select(one=True)['name'], 'test_delete')
        b1.delete()
        self.assertEqual(b1.select(one=True), None)

    def test_delete_bangumi(self):
        # Deprecated test

        # b1 = Bangumi(name='test_delete_all', update_time='Sun')
        # b2 = Bangumi(name='test_delete_all2', update_time='Sun')
        # Bangumi.delete_bangumi()
        # self.assertEqual(b1.select(one=True), None)
        # self.assertEqual(b2.select(one=True), None)
        pass


class FollowedTest(unittest.TestCase):
    def setUp(self):
        if not os.path.exists(DB_PATH):
            self.db = sqlite3.connect(DB_PATH)
            self.conn = self.db.cursor()
            self.conn.execute(CREATE_TABLE_BANGUMI)
            self.conn.execute(CREATE_TABLE_FOLLOWED)
            self.conn.execute(CREATE_TABLE_DOWNLOAD)
        else:
            self.db = sqlite3.connect(DB_PATH)
            self.conn = self.db.cursor()
        self.db.commit()

    def tearDown(self):
        self.db.close()
        os.remove(DB_PATH)

    def test_add_and_remove_followed(self):
        f = Followed(bangumi_name='test_add_and_remove_followed', status=STATUS_FOLLOWED, episode=6)
        f.save()
        b = Bangumi(name='test_add_and_remove_followed')
        b.save()
        bangumi_data = b.select(one=True, join='LEFT JOIN %s ON %s.bangumi_name=%s.name' % (Followed.table,
                                                                                            Followed.table,
                                                                                            Bangumi.table))
        self.assertEqual(bangumi_data['status'], STATUS_FOLLOWED)
        f.delete()
        bangumi_data = b.select(one=True, join='LEFT JOIN %s ON %s.bangumi_name=%s.name' % (Followed.table,
                                                                                            Followed.table,
                                                                                            Bangumi.table))
        self.assertEqual(bangumi_data['status'], STATUS_NORMAL)
